"""Provides 'Job' dataclass"""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Protocol

if TYPE_CHECKING:
    from croniter import croniter

    from nawah.types import NawahEnv


@dataclass(kw_only=True)
class Job:
    """Job dataclass serves role of defining items of 'jobs' Config Attr which
    are callabled, called periodically per cron-based schedule"""

    job: "JobCallable"
    schedule: str
    prevent_disable: bool

    @property
    def cron_schedule(self) -> "croniter":
        """Getter for '_cron_schedule' attr"""
        return self._cron_schedule

    @cron_schedule.setter
    def cron_schedule(self, val: "croniter"):
        """Setter for '_cron_schedule' attr"""
        self._cron_schedule = val

    @property
    def next_time(self) -> str:
        """Getter for '_next_time' attr"""
        return self._next_time

    @next_time.setter
    def next_time(self, val: str):
        """Setter for '_next_time' attr"""
        self._next_time = val

    @property
    def disabled(self) -> bool:
        """Getter for '_disabled' attr"""
        return self._disabled

    @disabled.setter
    def disabled(self, val: bool):
        """Setter for '_disabled' attr"""
        self._disabled = val


class JobCallable(Protocol):
    """Provides type-hint for 'job' callable of 'Job'"""

    # pylint: disable=too-few-public-methods
    def __call__(
        self,
        env: "NawahEnv",
    ) -> None:
        ...



"""Provides exceptions used in Nawah"""

from ._attr import (
    InvalidAttrTypeArgException,
    InvalidAttrTypeException,
    InvalidAttrTypeRefException,
    JSONPathNotFoundException,
)
from ._call import (
    InvalidCallEndpointException,
    InvalidFuncException,
    InvalidModuleException,
    NotPermittedException,
)
from ._config import ConfigException
from ._data import UnknownDeleteStrategyException
from ._func import (
    FuncException,
    InvalidDocAttrException,
    InvalidDocException,
    InvalidQueryAttrException,
    InvalidQueryException,
    MissingDocAttrException,
    MissingQueryAttrException,
)
from ._query import (
    InvalidQueryOperTypeException,
    InvalidQueryStepAttrLenException,
    InvalidQueryStepAttrTypeException,
    InvalidQueryStepLenException,
    InvalidQueryStepTypeException,
    UnknownQueryOperException,
)
from ._validate import InvalidAttrException, MissingAttrException
from ._var import InvalidLocaleException, InvalidVarException

__all__ = [
    "InvalidAttrTypeArgException",
    "InvalidAttrTypeException",
    "InvalidAttrTypeRefException",
    "JSONPathNotFoundException",
    "InvalidCallEndpointException",
    "InvalidFuncException",
    "InvalidModuleException",
    "NotPermittedException",
    "ConfigException",
    "UnknownDeleteStrategyException",
    "FuncException",
    "InvalidDocAttrException",
    "InvalidDocException",
    "InvalidQueryAttrException",
    "InvalidQueryException",
    "MissingDocAttrException",
    "MissingQueryAttrException",
    "InvalidQueryOperTypeException",
    "InvalidQueryStepAttrLenException",
    "InvalidQueryStepAttrTypeException",
    "InvalidQueryStepLenException",
    "InvalidQueryStepTypeException",
    "UnknownQueryOperException",
    "InvalidAttrException",
    "MissingAttrException",
    "InvalidLocaleException",
    "InvalidVarException",
]



"""Provides Exception classes for Base Functions"""

# START: Shared exceptions
class UtilityModuleDataCallException(Exception):
    """Raised if a data call was made for a Utility Module"""

    status = 400

    def __init__(self, *, module_name, func_name):
        super().__init__(
            UtilityModuleDataCallException.format_msg(
                module_name=module_name, func_name=func_name
            ),
            {"module_name": module_name, "func_name": func_name},
        )

    @staticmethod
    def format_msg(*, module_name, func_name):
        """Formats exception message"""

        return (
            f"Nawah Function '{func_name}' can't be called on Nawah Utility module "
            f"'{module_name}'"
        )


class DuplicateUniqueException(Exception):
    """Raised if 'create', 'update' call has at least one 'unique_attr' value in 'doc'
    that matches another doc"""

    status = 400

    def __init__(self, *, unique_attrs):
        super().__init__(
            DuplicateUniqueException.format_msg(unique_attrs=unique_attrs),
            {"unique_attrs": unique_attrs},
        )

    @staticmethod
    def format_msg(*, unique_attrs):
        """Formats exception message"""

        return f"A doc matching at least one of '{unique_attrs}' already exists"


class InvalidDocException(Exception):
    """Raised if 'delete_file', 'retrieve_file' call 'query' does not return any doc"""

    status = 400

    def __init__(self, *, doc_id):
        super().__init__(
            InvalidDocException.format_msg(doc_id=doc_id), {"doc_id": doc_id}
        )

    @staticmethod
    def format_msg(*, doc_id):
        """Formats exception message"""

        return f"Query for doc with '_id' '{doc_id}' returned no results"


# END: Shared exceptions

# START: create exceptions
class NoDocCreatedException(Exception):
    """Raised if 'create' call creates no doc"""

    status = 400

    def __init__(self, *, module_name):
        super().__init__(
            NoDocCreatedException.format_msg(module_name=module_name),
            {"module_name": module_name},
        )

    @staticmethod
    def format_msg(*, module_name):
        """Formats exception message"""

        return f"No documents were created for module '{module_name}'"


# END: create exceptions

# START: delete exceptions
class NoDocDeletedException(Exception):
    """Raised if 'delete' call deletes no doc"""

    status = 400

    def __init__(self, *, module_name):
        super().__init__(
            NoDocDeletedException.format_msg(module_name=module_name),
            {"module_name": module_name},
        )

    @staticmethod
    def format_msg(*, module_name):
        """Formats exception message"""

        return f"No documents were deleted for module '{module_name}'"


# END: delete exceptions

# START: read exceptions
class NoDocFoundException(Exception):
    """Raised if 'read' call founds no doc"""

    status = 400

    def __init__(self, *, module_name):
        super().__init__(
            NoDocFoundException.format_msg(module_name=module_name),
            {"module_name": module_name},
        )

    @staticmethod
    def format_msg(*, module_name):
        """Formats exception message"""

        return f"No documents were found for module '{module_name}'"


# END: read exceptions

# START: delete_file exceptions
class InvalidDeleteFileDocAttrException(Exception):
    """Raised if 'delete_file' call 'query' returns doc with attr not valid for operation"""

    status = 400

    def __init__(self, *, doc_id, attr_name):
        super().__init__(
            InvalidDeleteFileDocAttrException.format_msg(
                doc_id=doc_id, attr_name=attr_name
            ),
            {"doc_id": doc_id, "attr_name": attr_name},
        )

    @staticmethod
    def format_msg(*, doc_id, attr_name):
        """Formats exception message"""

        return f"Query for doc with '_id' '{doc_id}' returned doc with invalid value for attr '{attr_name}'"


class InvalidDeleteFileIndexException(Exception):
    """Raised if 'delete_file' call 'query' returns doc with index for call not valid for operation"""

    status = 400

    def __init__(self, *, doc_id, attr_name, index):
        super().__init__(
            InvalidDeleteFileIndexException.format_msg(
                doc_id=doc_id, attr_name=attr_name, index=index
            ),
            {"doc_id": doc_id, "attr_name": attr_name, "index": index},
        )

    @staticmethod
    def format_msg(*, doc_id, attr_name, index):
        """Formats exception message"""

        return f"Index '{index}' for attr '{attr_name}' for doc with '_id' '{doc_id}' is invalid"


class InvalidDeleteFileIndexValueException(Exception):
    """Raised if 'delete_file' call 'query' returns doc with index value for call not valid for operation"""

    status = 400

    def __init__(self, *, doc_id, attr_name, index, index_val_type):
        super().__init__(
            InvalidDeleteFileIndexValueException.format_msg(
                doc_id=doc_id,
                attr_name=attr_name,
                index=index,
                index_val_type=index_val_type,
            ),
            {
                "doc_id": doc_id,
                "attr_name": attr_name,
                "index": index,
                "index_val_type": index_val_type,
            },
        )

    @staticmethod
    def format_msg(*, doc_id, attr_name, index, index_val_type):
        """Formats exception message"""

        return f"Index '{index}' for attr '{attr_name}' for doc with '_id' '{doc_id}' is of invalid type '{index_val_type}'"


class InvalidDeleteFileMismatchNameException(Exception):
    """Raised if 'delete_file' call 'query' returns doc with index value file name not matching 'query'"""

    status = 400

    def __init__(self, *, doc_id, attr_name, index, query_file_name, index_file_name):
        super().__init__(
            InvalidDeleteFileMismatchNameException.format_msg(
                doc_id=doc_id,
                attr_name=attr_name,
                index=index,
                query_file_name=query_file_name,
                index_file_name=index_file_name,
            ),
            {
                "doc_id": doc_id,
                "attr_name": attr_name,
                "index": index,
                "query_file_name": query_file_name,
                "index_file_name": index_file_name,
            },
        )

    @staticmethod
    def format_msg(*, doc_id, attr_name, index, query_file_name, index_file_name):
        """Formats exception message"""

        return f"Index '{index}' for attr '{attr_name}' for doc with '_id' '{doc_id}' is for file '{index_file_name}', not '{query_file_name}'"


# END: delete_file exceptions

# START: obtain_lock exceptions
class FailedObtainLockException(Exception):
    """Raised if 'obtain_lock' call failed to obtain lock"""

    status = 400

    def __init__(self, *, module_name):
        super().__init__(
            FailedObtainLockException.format_msg(module_name=module_name),
            {"module_name": module_name},
        )

    @staticmethod
    def format_msg(*, module_name):
        """Formats exception message"""

        return f"Failed to obtain lock for '{module_name}'"


# END: obtain_lock exceptions

# START: delete_lock exceptions
class FailedDeleteLockException(Exception):
    """Raised if 'obtain_lock','delete_lock' call failed to deleted obtaiend lock"""

    status = 500

    def __init__(self, *, module_name, lock_id):
        super().__init__(
            FailedDeleteLockException.format_msg(
                module_name=module_name, lock_id=lock_id
            ),
            {"module_name": module_name, "lock_id": lock_id},
        )

    @staticmethod
    def format_msg(*, module_name, lock_id):
        """Formats exception message"""

        return f"Failed to delete lock '{lock_id}' for '{module_name}'. Delete manually now"


# END: delete_lock exceptions

# START: update exceptions
class UpdateMultiUniqueException(Exception):
    """Raised if 'update' call is attempted on multiple docs
    where at least one 'unique_attr' is present in 'update_doc'"""

    status = 400

    def __init__(self):
        super().__init__(UpdateMultiUniqueException.format_msg())

    @staticmethod
    def format_msg():
        """Formats exception message"""

        return (
            "Update call query has more than one doc as results. This would result in"
            "duplication"
        )


class EmptyUpdateDocException(Exception):
    """Raised if 'update' call has empty \'doc\'"""

    status = 400

    def __init__(self, *, module_name):
        super().__init__(
            EmptyUpdateDocException.format_msg(module_name=module_name),
            {"module_name": module_name},
        )

    @staticmethod
    def format_msg(*, module_name):
        """Formats exception message"""

        return f"Update 'doc' is empty for module '{module_name}'"


class NoDocUpdatedException(Exception):
    """Raised if 'update' call updates no doc"""

    status = 400

    def __init__(self, *, module_name):
        super().__init__(
            NoDocUpdatedException.format_msg(module_name=module_name),
            {"module_name": module_name},
        )

    @staticmethod
    def format_msg(*, module_name):
        """Formats exception message"""

        return f"No documents were updated for module '{module_name}'"


# END: update exceptions

# START: retrieve_file exceptions
class FileNotFoundException(Exception):
    """Raised by 'retrieve_file' if failed to locate attr, or attr does not match requested file name"""

    status = 400

    def __init__(self, *, doc_id, attr_name, file_name):
        super().__init__(
            FileNotFoundException.format_msg(
                doc_id=doc_id, attr_name=attr_name, file_name=file_name
            ),
            {"doc_id": doc_id, "attr_name": attr_name, "file_name": file_name},
        )

    @staticmethod
    def format_msg(*, doc_id, attr_name, file_name):
        """Formats exception message"""

        return f"Failed to find file for doc '_id' '{doc_id}', attr '{attr_name}', file '{file_name}'"


class FileNotImageException(Exception):
    """Raised by 'retrieve_file' if attemtped to generate thumbnail of non-image file"""

    status = 400

    def __init__(self, *, doc_id, attr_name, file_name, file_type):
        super().__init__(
            FileNotImageException.format_msg(
                doc_id=doc_id,
                attr_name=attr_name,
                file_name=file_name,
                file_type=file_type,
            ),
            {
                "doc_id": doc_id,
                "attr_name": attr_name,
                "file_name": file_name,
                "file_type": file_type,
            },
        )

    @staticmethod
    def format_msg(*, doc_id, attr_name, file_name, file_type):
        """Formats exception message"""

        return f"Can't generate thumbnail for file for doc '_id' '{doc_id}', attr '{attr_name}', file '{file_name}', of type '{file_type}'"


# END: retrieve_file exceptions



import datetime
import json
import logging
import re
from typing import TYPE_CHECKING, MutableMapping, MutableSequence, cast

import aiohttp.web
from multidict import MultiDict
from passlib.hash import pbkdf2_sha512
from requests_toolbelt.multipart import decoder

from nawah import data as Data
from nawah.classes import app_encoder
from nawah.config import Config
from nawah.enums import Event
from nawah.exceptions import InvalidAttrException

from .._call import call
from .._config import _compile_anon_session, _compile_anon_user
from .._validate import validate_doc

if TYPE_CHECKING:
    from nawah.classes import Attr
    from nawah.types import NawahDoc, NawahEnv

logger = logging.getLogger("nawah")


async def _http_handler(request: aiohttp.web.Request):
    headers = MultiDict(
        [
            ("Server", "Nawah"),
            ("Powered-By", "Nawah, https://nawah.masaar.com"),
            ("Access-Control-Allow-Origin", "*"),
            ("Access-Control-Allow-Methods", "GET,POST,OPTIONS"),
            (
                "Access-Control-Allow-Headers",
                "Content-Type,X-Auth-Bearer,X-Auth-Token,X-Auth-App",
            ),
            ("Access-Control-Expose-Headers", "Content-Disposition"),
        ]
    )

    logger.debug("Received new %s request: %s", request.method, request.match_info)

    if request.method == "OPTIONS":
        return aiohttp.web.Response(
            status=200,
            headers=headers,
            body=app_encoder.encode(
                {
                    "status": 200,
                    "msg": "OPTIONS request is allowed",
                }
            ),
        )

    # Check for IP quota
    if str(request.remote) not in Config.sys.ip_quota:
        Config.sys.ip_quota[str(request.remote)] = {
            "counter": Config.quota_ip_min,
            "last_check": datetime.datetime.utcnow(),
        }
    else:
        if (
            datetime.datetime.utcnow()
            - Config.sys.ip_quota[str(request.remote)]["last_check"]
        ).seconds > 259:
            Config.sys.ip_quota[str(request.remote)][
                "last_check"
            ] = datetime.datetime.utcnow()
            Config.sys.ip_quota[str(request.remote)]["counter"] = Config.quota_ip_min
        else:
            if Config.sys.ip_quota[str(request.remote)]["counter"] - 1 <= 0:
                logger.warning(
                    "Denying '%s' request from '%s' for hitting IP quota",
                    request.method,
                    request.remote,
                )
                headers["Content-Type"] = "application/json; charset=utf-8"
                return aiohttp.web.Response(
                    status=429,
                    headers=headers,
                    body=app_encoder.encode(
                        {
                            "status": 429,
                            "msg": "You have hit calls quota from this IP",
                            "args": {"code": "CORE_REQ_IP_QUOTA_HIT"},
                        }
                    ),
                )

            Config.sys.ip_quota[str(request.remote)]["counter"] -= 1

    module_name = request.url.parts[1]
    func_name = request.url.parts[2]
    request_args = dict(request.match_info.items())

    # Extract Args Sets based on request.method
    args_sets = Config.modules[module_name].funcs[func_name].query_attrs
    args_sets = cast(MutableSequence[MutableMapping[str, "Attr"]], args_sets)

    # Attempt to validate query as doc
    for args_set in args_sets or []:
        if len(args_set.keys()) == len(args_set.keys()) and sum(
            1 for arg in args_set.keys() if arg in args_set.keys()
        ) == len(args_set.keys()):
            # Check presence and validate all attrs in doc args
            try:
                exception: Exception
                validate_doc(mode="create", doc=request_args, attrs=args_set)  # type: ignore
            except InvalidAttrException as e:
                exception = e
                headers["Content-Type"] = "application/json; charset=utf-8"
                return aiohttp.web.Response(
                    status=400,
                    headers=headers,
                    body=app_encoder.encode(
                        {
                            "status": 400,
                            "msg": f"{str(e)} for '{request.method}' request on module "
                            "'{module_name}'",
                            "args": {"code": f"{module_name.upper()}_INVALID_ATTR"},
                        }
                    ).encode("utf-8"),
                )
            break

    conn = Data.create_conn()
    env: "NawahEnv" = {
        "conn": conn,
        "REMOTE_ADDR": request.remote or "localhost",
        "client_app": "__public",
        "args": {},
    }

    try:
        env["HTTP_USER_AGENT"] = request.headers["user-agent"]
        env["HTTP_ORIGIN"] = request.headers["origin"]
    except:
        env["HTTP_USER_AGENT"] = ""
        env["HTTP_ORIGIN"] = ""

    if "X-Auth-Bearer" in request.headers or "X-Auth-Token" in request.headers:
        logger.debug("Detected 'X-Auth' header[s]")
        if (
            "X-Auth-Bearer" not in request.headers
            or "X-Auth-Token" not in request.headers
            or "X-Auth-App" not in request.headers
        ):
            logger.debug("Denying request due to missing 'X-Auth' header")
            headers["Content-Type"] = "application/json; charset=utf-8"
            return aiohttp.web.Response(
                status=400,
                headers=headers,
                body=app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "One 'X-Auth' headers was set but not the other",
                    }
                ).encode("utf-8"),
            )
        # [TODO] Add condition to now allow a connection from a client with client_app = __sys
        if Config.client_apps and (
            request.headers["X-Auth-App"] not in Config.client_apps
            or (
                Config.client_apps[request.headers["X-Auth-App"]].type == "web"
                and env["HTTP_ORIGIN"]
                not in Config.client_apps[request.headers["X-Auth-App"]].origin
            )
        ):
            logger.debug("Denying request due to unauthorised client_app")
            headers["Content-Type"] = "application/json; charset=utf-8"
            return aiohttp.web.Response(
                status=403,
                headers=headers,
                body=app_encoder.encode(
                    {
                        "status": 403,
                        "msg": "X-Auth headers could not be verified",
                        "args": {"code": "CORE_SESSION_INVALID_XAUTH"},
                    }
                ).encode("utf-8"),
            )
        try:
            session_results = await call(
                "session/read",
                skip_events=[Event.PERM],
                env=env,
                query=[
                    {
                        "_id": request.headers["X-Auth-Bearer"],
                    }
                ],
            )
        except:
            headers["Content-Type"] = "application/json; charset=utf-8"
            if Config.debug:
                return aiohttp.web.Response(
                    status=500,
                    headers=headers,
                    body=app_encoder.encode(
                        {
                            "status": 500,
                            "msg": f"Unexpected error has occurred [{str(exception)}]",
                            "args": {
                                "code": "CORE_SERVER_ERROR",
                                "err": str(exception),
                            },
                        }
                    ).encode("utf-8"),
                )

            return aiohttp.web.Response(
                status=500,
                headers=headers,
                body=app_encoder.encode(
                    {
                        "status": 500,
                        "msg": "Unexpected error has occurred",
                        "args": {"code": "CORE_SERVER_ERROR"},
                    }
                ).encode("utf-8"),
            )

        if not session_results["args"]["count"] or not pbkdf2_sha512.verify(
            request.headers["X-Auth-Token"],
            session_results["args"]["docs"][0]["token_hash"],
        ):
            logger.debug("Denying request due to missing failed Call Authorisation")
            headers["Content-Type"] = "application/json; charset=utf-8"
            return aiohttp.web.Response(
                status=403,
                headers=headers,
                body=app_encoder.encode(
                    {
                        "status": 403,
                        "msg": "X-Auth headers could not be verified",
                        "args": {"code": "CORE_SESSION_INVALID_XAUTH"},
                    }
                ).encode("utf-8"),
            )
        session = session_results["args"]["docs"][0]
        session_results = await call(
            "session/reauth",
            skip_events=[Event.PERM],
            env=env,
            query=[
                {
                    "_id": request.headers["X-Auth-Bearer"],
                    "token": request.headers["X-Auth-Token"],
                    "groups": [],
                }
            ],
        )
        logger.debug("Denying request due to fail to reauth")
        if session_results["status"] != 200:
            headers["Content-Type"] = "application/json; charset=utf-8"
            return aiohttp.web.Response(
                status=403,
                headers=headers,
                body=app_encoder.encode(session_results).encode("utf-8"),
            )

        session = session_results["args"]["session"]
    else:
        anon_user = _compile_anon_user()
        anon_session = _compile_anon_session()
        anon_session["user"] = anon_user
        session = anon_session

    env["session"] = json.loads(app_encoder.encode(session))

    doc: "NawahDoc" = {}

    if "Content-Type" in request.headers:
        doc_content = await request.content.read()
        # Check Content-Type to decide how to deserialise content
        if request.headers["Content-Type"].startswith("multipart/form-data;"):
            multipart_content_type = request.headers["Content-Type"]
            doc = {}
            for part in decoder.MultipartDecoder(
                doc_content, multipart_content_type
            ).parts:
                content_disposition = part.headers[b"Content-Disposition"].decode(
                    "UTF-8"
                )
                attr_name_match = re.findall('name="([^"]+)"', content_disposition)
                if not attr_name_match:
                    continue
                attr_name = attr_name_match[0]
                doc[attr_name] = part.content
                # For non-file values, decode unto str
                if "filename" not in content_disposition:
                    doc[attr_name] = doc[attr_name].decode("UTF-8")
        elif request.headers["Content-Type"].startswith("application/json;"):
            doc = json.loads(doc_content)

    results = await call(
        f"{module_name}/{func_name}", env=env, query=[request_args], doc=doc
    )

    logger.debug("Closing connection")
    env["conn"].close()

    if "return" not in results["args"] or results["args"]["return"] == "json":
        if "return" in results["args"]:
            del results["args"]["return"]
        headers["Content-Type"] = "application/json; charset=utf-8"
        if results["status"] == 404:
            return aiohttp.web.Response(
                status=results["status"],
                headers=headers,
                body=app_encoder.encode(
                    {"status": 404, "msg": "Requested content not found"}
                ).encode("utf-8"),
            )
        return aiohttp.web.Response(
            status=results["status"],
            headers=headers,
            body=app_encoder.encode(results),
        )

    if results["args"]["return"] == "file":
        del results["args"]["return"]
        status = results["status"]
        expiry_time = datetime.datetime.utcnow() + datetime.timedelta(days=30)
        headers["Last-Modified"] = str(results["args"]["docs"][0]["lastModified"])
        headers["Content-Type"] = results["args"]["docs"][0]["type"]
        headers["Cache-Control"] = "public, max-age=31536000"
        headers["Expires"] = expiry_time.strftime("%a, %d %b %Y %H:%M:%S GMT")
        content = results["args"]["docs"][0]["content"]
        if "If-Modified-Since" in request.headers:
            if request.headers["If-Modified-Since"] == headers["Last-Modified"]:
                status = 304
                content = b""
        return aiohttp.web.Response(
            status=status,
            headers=headers,
            body=content,
        )

    if results["args"]["return"] == "msg":
        del results["args"]["return"]
        headers["Content-Type"] = "application/json; charset=utf-8"
        return aiohttp.web.Response(
            status=results["status"], headers=headers, body=results["msg"]
        )

    headers["Content-Type"] = "application/json; charset=utf-8"
    return aiohttp.web.Response(
        status=405,
        headers=headers,
        body=app_encoder.encode({"status": 405, "msg": "METHOD NOT ALLOWED"}),
    )



"""Provides exceptions related to Nawah Function"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..types import ResultsArgs


class FuncException(Exception):
    """Raised by \'call\' Utility when encounters an exception raised
    by Nawah Function callable"""

    def __init__(self, *, status: int, msg: str, args: "ResultsArgs"):
        if "code" not in args:
            raise Exception(
                "Invalid value for 'args' for FuncException. Missing item 'code'"
            )

        super().__init__(msg, {"status": status, "args": args})


class MissingQueryAttrException(Exception):
    """Raied by '_check_query_attrs' if 'attr_name' is missing from checked Query object"""

    status = 400

    def __init__(self, *, attr_name):
        super().__init__(
            MissingQueryAttrException.format_msg(attr_name=attr_name),
            {"attr_name": attr_name},
        )

    @staticmethod
    def format_msg(*, attr_name):
        """Formats exception message"""

        return f"Missing attr '{attr_name}' from Query"


class InvalidQueryAttrException(Exception):
    """Raised by '_check_query_attrs' if 'attr' has invalid value"""

    status = 400

    def __init__(self, *, attr_name, attr_type, val_type):
        super().__init__(
            InvalidQueryAttrException.format_msg(
                attr_name=attr_name, attr_type=attr_type, val_type=val_type
            ),
            {"attr_name": attr_name, "attr_type": attr_type, "val_type": val_type},
        )

    @staticmethod
    def format_msg(*, attr_name, attr_type, val_type):
        """Formats exception message"""

        return (
            f"Invalid attr '{attr_name}' of type '{val_type}' with required type "
            f"'{attr_type}'"
        )


class InvalidQueryException(Exception):
    """Raised by '_check_query_attrs', when fail to match all 'query_attrs' sets"""

    status = 400

    def __init__(self, *, query_attrs_sets):
        super().__init__(
            InvalidQueryException.format_msg(query_attrs_sets=query_attrs_sets),
            {"query_attrs_sets": query_attrs_sets},
        )

    @staticmethod
    def format_msg(*, query_attrs_sets):
        """Formats exception message"""

        return f"Invalid Query. Matched against: {query_attrs_sets}"


class MissingDocAttrException(Exception):
    """Raied by '_check_doc_attrs' if 'attr_name' is missing from checked Doc object"""

    status = 400

    def __init__(self, *, attr_name):
        super().__init__(
            MissingDocAttrException.format_msg(attr_name=attr_name),
            {"attr_name": attr_name},
        )

    @staticmethod
    def format_msg(*, attr_name):
        """Formats exception message"""

        return f"Missing attr '{attr_name}' from Doc"


class InvalidDocAttrException(Exception):
    """Raised by '_check_doc_attrs' if 'attr' has invalid value"""

    status = 400

    def __init__(self, *, attr_name, attr_type, val_type):
        super().__init__(
            InvalidDocAttrException.format_msg(
                attr_name=attr_name, attr_type=attr_type, val_type=val_type
            ),
            {"attr_name": attr_name, "attr_type": attr_type, "val_type": val_type},
        )

    @staticmethod
    def format_msg(*, attr_name, attr_type, val_type):
        """Formats exception message"""

        return (
            f"Invalid attr '{attr_name}' of type '{val_type}' with required type "
            f"'{attr_type}'"
        )


class InvalidDocException(Exception):
    """Raised by '_check_doc_attrs', when fail to match all 'doc_attrs' sets"""

    status = 400

    def __init__(self, *, doc_attrs_sets):
        super().__init__(
            InvalidDocException.format_msg(doc_attrs_sets=doc_attrs_sets),
            {"doc_attrs_sets": doc_attrs_sets},
        )

    @staticmethod
    def format_msg(*, doc_attrs_sets):
        """Formats exception message"""

        return f"Invalid Doc. Matched against: {doc_attrs_sets}"



"""Provides types used in Nawah"""

from ._types import (
    AnalyticsEvents,
    AppPackage,
    IPQuota,
    NawahDoc,
    NawahEnv,
    NawahEvents,
    NawahQueryOperAnd,
    NawahQueryOperOr,
    NawahQuerySpecial,
    NawahQuerySpecialGroup,
    NawahQueryStep,
    NawahSession,
    Results,
    ResultsArgs,
)

__all__ = [
    "AnalyticsEvents",
    "AppPackage",
    "IPQuota",
    "NawahDoc",
    "NawahEnv",
    "NawahEvents",
    "NawahQueryOperAnd",
    "NawahQueryOperOr",
    "NawahQuerySpecial",
    "NawahQuerySpecialGroup",
    "NawahQueryStep",
    "NawahSession",
    "Results",
    "ResultsArgs",
]



"""Provides loops to keep asyncio running, running minutely checks"""

import asyncio
import logging
import sys
import time
from typing import TYPE_CHECKING

import aiohttp.web

from nawah.config import Config

from ._common_handlers import (_not_allowed_handler, _not_found_handler,
                               _root_handler)
from ._http_handler import _http_handler
from ._utils import (_check_ips_quota, _check_sessions, _execute_jobs,
                     _populate_routes)
from ._websocket_handler import _websocket_handler

if TYPE_CHECKING:
    from aiohttp.web import Application

logger = logging.getLogger("nawah")


async def _jobs_loop():
    # Connection Timeout Workflow
    logger.debug("Time to check for sessions!")
    logger.debug("Current sessions: %s", Config.sys.sessions)
    session_ids = list(Config.sys.sessions)
    asyncio.create_task(_check_sessions(session_ids))

    # Calls Quota Workflow - Clean-up Sequence
    logger.debug("Time to check for IPs quotas!")
    ips = list(Config.sys.ip_quota)
    asyncio.create_task(_check_ips_quota(ips))

    # Jobs Workflow
    logger.debug("Time to check for jobs!")
    asyncio.create_task(_execute_jobs())

    # [TODO] Re-implement
    # try:
    #     logger.debug('Time to check for files timeout!')
    #     files_task = asyncio.create_task(
    #         call(
    #             'file/delete',
    #             skip_events=[Event.PERM],
    #             env=Config.sys.env,
    #             query=[
    #                 {
    #                     'create_time': {
    #                         '$lt': (
    #                             datetime.datetime.utcnow()
    #                             - datetime.timedelta(seconds=Config.file_upload_timeout)
    #                         ).isoformat()
    #                     }
    #                 }
    #             ],
    #         )
    #     )
    #     # logger.debug('Files timeout results:')
    #     # logger.debug('-status: %s', files_results['status'])
    #     # logger.debug('-msg: %s', files_results['msg'])
    #     # logger.debug('-args.docs: %s', files_results["args"]['docs'])
    # except Exception:
    #     logger.error('An error occurred. Details: %s', traceback.format_exc())


def _create_error_middleware(overrides):
    @aiohttp.web.middleware
    async def error_middleware(request, handler):
        try:
            response = await handler(request)
            override = overrides.get(response.status)
            if override:
                return await override(request)
            return response
        except aiohttp.web.HTTPException as ex:
            override = overrides.get(ex.status)
            if override:
                return await override(request)
            raise

    return error_middleware


async def _web_loop(app: "Application", /):
    runner = aiohttp.web.AppRunner(app)
    await runner.setup()
    site = aiohttp.web.TCPSite(runner, "0.0.0.0", Config.port)
    await site.start()
    logger.info("Serving on 0.0.0.0:%s", Config.port)
    while True:
        await asyncio.sleep(60)
        asyncio.create_task(_jobs_loop())


def _run_app():
    get_routes, post_routes = _populate_routes()

    logger.debug(
        "Loaded modules: %s",
        {module_name: module.attrs for module_name, module in Config.modules.items()},
    )
    logger.debug(
        "Config has attrs: %s",
        {
            k: str(v)
            for k, v in Config.__dict__.items()
            if not isinstance(v, classmethod) and not k.startswith("_")
        },
    )
    logger.debug("Generated get_routes: %s", get_routes)
    logger.debug("Generated post_routes: %s", post_routes)

    app = aiohttp.web.Application()
    app.middlewares.append(
        _create_error_middleware(
            {
                404: _not_found_handler,
                405: _not_allowed_handler,
            }
        )
    )
    app.router.add_route("GET", "/", _root_handler)
    app.router.add_route("*", "/ws", _websocket_handler)
    for route in get_routes:
        app.router.add_route("GET", route, _http_handler)
    for route in post_routes:
        app.router.add_route("POST", route, _http_handler)
        app.router.add_route("OPTIONS", route, _http_handler)
    logger.info("Welcome to Nawah")

    try:
        asyncio.run(_web_loop(app))
    except KeyboardInterrupt:
        if time.localtime().tm_hour >= 21 or time.localtime().tm_hour <= 4:
            msg = "night"
        elif time.localtime().tm_hour >= 18:
            msg = "evening"
        elif time.localtime().tm_hour >= 12:
            msg = "afternoon"
        elif time.localtime().tm_hour >= 5:
            msg = "morning"

        logger.info("Have a great %s!", msg)

        sys.exit()



import asyncio
import datetime
import json
import logging
import random
import traceback
from functools import partial
from typing import TYPE_CHECKING

import aiohttp.web

from nawah import data as Data
from nawah.classes import app_encoder
from nawah.config import Config

from .._call import call
from .._config import _compile_anon_session, _compile_anon_user
from .._val import camel_to_upper
from ._utils import _close_session

if TYPE_CHECKING:
    from asyncio.futures import Future

    from nawah.types import NawahEnv

logger = logging.getLogger("nawah")


async def _websocket_handler(request: aiohttp.web.Request):
    conn = Data.create_conn()
    logger.debug("Websocket connection starting with client at '%s'", request.remote)
    ws = aiohttp.web.WebSocketResponse()
    await ws.prepare(request)

    session_lock_key = random.random()

    while Config.sys.session_lock != session_lock_key:
        if Config.sys.session_lock is None:
            Config.sys.session_lock = session_lock_key

    Config.sys.session_counter += 1

    env: "NawahEnv" = {
        "id": str(Config.sys.session_counter),
        "conn": conn,
        "REMOTE_ADDR": request.remote or "localhost",
        "ws": ws,
        "init": False,
        "last_call": datetime.datetime.utcnow(),
        "quota": {
            "counter": Config.quota_anon_min,
            "last_check": datetime.datetime.utcnow(),
        },
        "args": {},
    }
    Config.sys.sessions[str(Config.sys.session_counter)] = env
    Config.sys.session_lock = None

    try:
        env["HTTP_USER_AGENT"] = request.headers["user-agent"]
        env["HTTP_ORIGIN"] = request.headers["origin"]
    except:
        env["HTTP_USER_AGENT"] = ""
        env["HTTP_ORIGIN"] = ""

    logger.debug(
        "Websocket connection #'%s' ready with client at '%s'",
        env["id"],
        env["REMOTE_ADDR"],
    )

    await ws.send_str(
        app_encoder.encode(
            {
                "status": 200,
                "msg": "Connection ready",
                "args": {"code": "CORE_CONN_READY"},
            }
        )
    )

    async for msg in ws:
        if "conn" not in env:
            await ws.close()
            break
        logger.debug(
            "Received new message from session #'%s': %s", env["id"], msg.data[:256]
        )
        if msg.type == aiohttp.WSMsgType.TEXT:
            logger.debug(
                "Config.sys.ip_quota on session #'%s': %s",
                env["id"],
                Config.sys.ip_quota,
            )
            logger.debug("session_quota: on session #'%s': %s", env["id"], env["quota"])
            # Check for IP quota
            if str(request.remote) not in Config.sys.ip_quota:
                Config.sys.ip_quota[str(request.remote)] = {
                    "counter": Config.quota_ip_min,
                    "last_check": datetime.datetime.utcnow(),
                }
            else:
                if (
                    datetime.datetime.utcnow()
                    - Config.sys.ip_quota[str(request.remote)]["last_check"]
                ).seconds > 59:
                    Config.sys.ip_quota[str(request.remote)][
                        "last_check"
                    ] = datetime.datetime.utcnow()
                    Config.sys.ip_quota[str(request.remote)][
                        "counter"
                    ] = Config.quota_ip_min
                else:
                    # If hit quota, deny call
                    if Config.sys.ip_quota[str(request.remote)]["counter"] - 1 <= 0:
                        logger.warning(
                            "Denying Websocket request from '%s' for hitting IP quota",
                            request.remote,
                        )
                        asyncio.create_task(
                            _handle_msg(
                                env=env,
                                msg=msg,
                                decline_quota="ip",
                            )
                        )
                        continue
                    # Else, update quota counter
                    Config.sys.ip_quota[str(request.remote)]["counter"] -= 1
            # Check for session quota
            if (datetime.datetime.utcnow() - env["quota"]["last_check"]).seconds > 59:
                env["quota"]["last_check"] = datetime.datetime.utcnow()
                env["quota"]["counter"] = (
                    (Config.quota_anon_min - 1)
                    if not env["session"]
                    or env["session"]["token"] == Config.anon_token
                    else (Config.quota_auth_min - 1)
                )
                asyncio.create_task(_handle_msg(env=env, msg=msg))
            else:
                # If hit quota, deny call
                if env["quota"]["counter"] - 1 <= 0:
                    asyncio.create_task(
                        _handle_msg(
                            env=env,
                            msg=msg,
                            decline_quota="session",
                        )
                    )
                    continue
                # Else, update quota counter
                env["quota"]["counter"] -= 1
                asyncio.create_task(_handle_msg(env=env, msg=msg))

    if "id" in env:
        await _close_session(env["id"])

    return ws


async def _handle_msg(
    env: "NawahEnv",
    msg: aiohttp.WSMessage,
    decline_quota: str = None,
):
    try:
        env["last_call"] = datetime.datetime.utcnow()
        try:
            env["session"]["token"]
        except Exception:
            anon_user = _compile_anon_user()
            anon_session = _compile_anon_session()
            anon_session["user"] = anon_user
            env["session"] = json.loads(app_encoder.encode(anon_session))
        res = json.loads(msg.data)

        # Check if msg should be denied for quota hit
        if decline_quota == "ip":
            await env["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 429,
                        "msg": "You have hit calls quota from this IP",
                        "args": {
                            "call_id": res["call_id"]
                            if "call_id" in res.keys()
                            else None,
                            "code": "CORE_REQ_IP_QUOTA_HIT",
                        },
                    }
                )
            )
            return

        if decline_quota == "session":
            await env["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 429,
                        "msg": "You have hit calls quota",
                        "args": {
                            "call_id": res["call_id"]
                            if "call_id" in res.keys()
                            else None,
                            "code": "CORE_REQ_SESSION_QUOTA_HIT",
                        },
                    }
                )
            )
            return

        logger.debug("Decoded request: %s", app_encoder.encode(res))

        if "endpoint" not in res.keys():
            await env["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "Request missing endpoint",
                        "args": {
                            "call_id": res["call_id"]
                            if "call_id" in res.keys()
                            else None,
                            "code": "CORE_REQ_NO_ENDPOINT",
                        },
                    }
                )
            )
            return

        if env["init"] is False:
            if res["endpoint"] != "conn/verify":
                await env["ws"].send_str(
                    app_encoder.encode(
                        {
                            "status": 1008,
                            "msg": "Request token is not accepted",
                            "args": {
                                "call_id": res["call_id"]
                                if "call_id" in res.keys()
                                else None,
                                "code": "CORE_REQ_NO_VERIFY",
                            },
                        }
                    )
                )
                await env["ws"].close()
                return

            # [TODO] Add condition to now allow a connection from a client with client_app = __sys
            if Config.client_apps and (
                "doc" not in res
                or "app" not in res["doc"]
                or res["doc"]["app"] not in Config.client_apps
                or (
                    Config.client_apps[res["doc"]["app"]].type == "web"
                    and env["HTTP_ORIGIN"]
                    not in Config.client_apps[res["doc"]["app"]].origin
                )
            ):
                await env["ws"].send_str(
                    app_encoder.encode(
                        {
                            "status": 1008,
                            "msg": "Request token is not accepted",
                            "args": {
                                "call_id": res["call_id"]
                                if "call_id" in res.keys()
                                else None,
                                "code": "CORE_REQ_NO_VERIFY",
                            },
                        }
                    )
                )
                await env["ws"].close()
                return

            env["init"] = True
            # [TODO] Add condition to now allow a connection from a client with client_app = __sys
            if not Config.client_apps:
                env["client_app"] = "__public"
            else:
                env["client_app"] = res["doc"]["app"]
            logger.debug("Connection on session #'%s' is verified", env["id"])
            # if Config.analytics_events['app_conn_verified']:
            #     asyncio.create_task(
            #         call(
            #             'analytic/create',
            #             skip_events=[Event.PERM],
            #             env=env,
            #             doc={
            #                 'event': 'CONN_VERIFIED',
            #                 'subevent': env['client_app'],
            #                 'args': {
            #                     'REMOTE_ADDR': env['REMOTE_ADDR'],
            #                     'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
            #                 },
            #             },
            #         )
            #     )
            await env["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 200,
                        "msg": "Connection established",
                        "args": {
                            "call_id": res["call_id"]
                            if "call_id" in res.keys()
                            else None,
                            "code": "CORE_CONN_OK",
                        },
                    }
                )
            )
            return

        if res["endpoint"] == "conn/close":
            logger.debug(
                f'Received connection close instructions on session #\'{env["id"]}\''
            )
            await env["ws"].close()
            return

        if res["endpoint"] == "heart/beat":
            logger.debug(f'Received connection heartbeat on session #\'{env["id"]}\'')
            await env["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 200,
                        "msg": "Heartbeat received",
                        "args": {
                            "call_id": res["call_id"]
                            if "call_id" in res.keys()
                            else None,
                            "code": "CORE_HEARTBEAT_OK",
                        },
                    }
                )
            )
            return

        res["endpoint"] = res["endpoint"]

        if (
            res["endpoint"] in ["session/auth", "session/reauth"]
            and str(env["session"]["_id"]) != "f00000000000000000000012"
        ):
            await env["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "You are already authed",
                        "args": {
                            "call_id": res["call_id"]
                            if "call_id" in res.keys()
                            else None,
                            "code": "CORE_SESSION_ALREADY_AUTHED",
                        },
                    }
                )
            )
            return

        if (
            res["endpoint"] == "session/signout"
            and str(env["session"]["_id"]) == "f00000000000000000000012"
        ):
            await env["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "Singout is not allowed for '__ANON' user",
                        "args": {
                            "call_id": res["call_id"]
                            if "call_id" in res.keys()
                            else None,
                            "code": "CORE_SESSION_ANON_SIGNOUT",
                        },
                    }
                )
            )
            return

        if "query" not in res.keys():
            res["query"] = []
        if "doc" not in res.keys():
            res["doc"] = {}
        if "call_id" not in res.keys():
            res["call_id"] = ""

        request = {
            "call_id": res["call_id"],
            # 'sid': res['sid'] or False,
            "query": res["query"],
            "doc": res["doc"],
            "path": res["endpoint"].split("/"),
        }

        if len(request["path"]) != 2:
            await env["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "Endpoint path is invalid",
                        "args": {
                            "call_id": request["call_id"],
                            "code": "CORE_REQ_INVALID_PATH",
                        },
                    }
                )
            )
            return

        module = request["path"][0]

        if module not in Config.modules:
            await env["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "Endpoint module is invalid",
                        "args": {
                            "call_id": request["call_id"],
                            "code": "CORE_REQ_INVALID_MODULE",
                        },
                    }
                )
            )
            return

        if request["path"][1] not in Config.modules[module].funcs:
            await env["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "Endpoint is invalid",
                        "args": {
                            "call_id": request["call_id"],
                            "code": "CORE_REQ_INVALID_METHOD",
                        },
                    }
                )
            )
            return

        if Config.modules[module].funcs[request["path"][1]].get_func:
            await env["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "Endpoint is a GET Function",
                        "args": {
                            "call_id": request["call_id"],
                            "code": "CORE_REQ_GET_FUNC",
                        },
                    }
                )
            )
            return

        # if not request['sid']:
        request["sid"] = "f00000000000000000000012"

        query = request["query"]
        doc = request["doc"]
        # await _process_file_obj(doc=doc, env=env)
        call_task = asyncio.create_task(
            call(
                f'{module}/{request["path"][1]}',
                skip_events=[],
                env=env,
                query=query,
                doc=doc,
            )
        )
        call_task.add_done_callback(
            partial(_call_callback, env=env, call_id=request["call_id"])
        )

    except Exception as e:
        logger.error(f"An error occurred. Details: {traceback.format_exc()}")
        if Config.debug:
            await env["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 500,
                        "msg": f"Unexpected error has occurred [{str(e)}]",
                        "args": {
                            "code": "CORE_SERVER_ERROR",
                            "err": str(e),
                            "call_id": request["call_id"],
                        },
                    }
                )
            )
        else:
            await env["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 500,
                        "msg": "Unexpected error has occurred",
                        "args": {
                            "code": "CORE_SERVER_ERROR",
                            "call_id": request["call_id"],
                        },
                    }
                )
            )


def _call_callback(call_task: "Future", env: "NawahEnv", call_id: str):
    try:
        results = call_task.result()
        results["args"]["call_id"] = call_id
        asyncio.create_task(env["ws"].send_str(app_encoder.encode(results)))
    except Exception as e:
        status = getattr(e, "status", 500)
        msg = "Unexpected error has occurred"
        code = camel_to_upper(e.__class__.__name__)
        if e.args:
            msg = e.args[0]
            if len(e.args) == 2:
                if isinstance(e.args[1], dict):
                    if "args" in e.args[1]:
                        if "code" in e.args[1]["args"]:
                            code = e.args[1]["args"]["code"]
                    if "status" in e.args[1]:
                        status = e.args[1]["status"]

        asyncio.create_task(
            env["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": status,
                        "msg": msg,
                        "args": {"code": code, "call_id": call_id},
                    }
                )
            )
        )



"""Provides 'user' Module Functions callables"""

from typing import TYPE_CHECKING, Optional

from bson import ObjectId

from nawah.config import Config
from nawah.enums import Event
from nawah.exceptions import FuncException
from nawah.utils import call

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahDoc, NawahEnv, NawahEvents, Results


async def _read(
    skip_events: "NawahEvents",
    env: "NawahEnv",
    query: "Query",
    skip_sanitise_results: Optional[bool],
) -> "Results":
    skip_events.append(Event.PERM)
    read_results = await call(
        "base/read", module_name="user", skip_events=skip_events, env=env, query=query
    )

    if read_results["status"] != 200:
        return read_results

    # Return results as is if both sanitise_results, user_settings are to be skipped
    if skip_sanitise_results is True:
        return read_results

    # Otherwise, iterate over results and apply not skipped action
    for i, _ in enumerate(read_results["args"]["docs"]):
        user = read_results["args"]["docs"][i]
        # Apply sanitise_results, removing hashes, if not skipped
        if skip_sanitise_results is not True:
            for user_attr_sanitise in Config.user_attrs_sanitise:
                del user[user_attr_sanitise]

    return read_results


async def _create(
    skip_events: "NawahEvents", env: "NawahEnv", doc: "NawahDoc"
) -> "Results":
    if Event.ATTRS_DOC not in skip_events:
        doc["groups"] = [ObjectId("f00000000000000000000013")]

    create_results = await call(
        "base/create", skip_events=[Event.PERM], module_name="user", env=env, doc=doc
    )

    return create_results


async def _read_privileges(env: "NawahEnv", query: "Query") -> "Results":
    # Confirm _id is valid
    results = await call(
        "user/read",
        skip_events=[Event.PERM],
        env=env,
        query=[{"_id": query["_id"][0]}],
        args={"skip_sanitise_results": True},
    )
    if not results["args"]["count"]:
        raise FuncException(
            status=400, msg="User is invalid.", args={"code": "INVALID_USER"}
        )
    user = results["args"]["docs"][0]
    for group in user["groups"]:
        group_results = await call(
            "group/read", skip_events=[Event.PERM], env=env, query=[{"_id": group}]
        )
        group = group_results["args"]["docs"][0]
        for privilege in group["privileges"].keys():
            if privilege not in user["privileges"].keys():
                user["privileges"][privilege] = []
            for i in range(len(group["privileges"][privilege])):
                if (
                    group["privileges"][privilege][i]
                    not in user["privileges"][privilege]
                ):
                    user["privileges"][privilege].append(
                        group["privileges"][privilege][i]
                    )
    return results


async def _add_group(
    skip_events: "NawahEvents", env: "NawahEnv", query: "Query", doc: "NawahDoc"
) -> "Results":
    # Check for list group attr
    if isinstance(doc["group"], list):
        for i in range(0, len(doc["group"]) - 1):
            await call(
                "user/add_group",
                skip_events=skip_events,
                env=env,
                query=query,
                doc={"group": doc["group"][i]},
            )
        doc["group"] = doc["group"][-1]
    # Confirm all basic args are provided
    doc["group"] = ObjectId(doc["group"])
    # Confirm group is valid
    results = await call(
        "group/read", skip_events=[Event.PERM], env=env, query=[{"_id": doc["group"]}]
    )
    if not results["args"]["count"]:
        raise FuncException(
            status=400, msg="Group is invalid.", args={"code": "INVALID_GROUP"}
        )
    # Get user details
    results = await call("user/read", skip_events=[Event.PERM], env=env, query=query)
    if not results["args"]["count"]:
        raise FuncException(
            status=400, msg="User is invalid.", args={"code": "INVALID_USER"}
        )
    user = results["args"]["docs"][0]
    # Confirm group was not added before
    if doc["group"] in user["groups"]:
        raise FuncException(
            status=400,
            msg="User is already a member of the group.",
            args={"code": "GROUP_ADDED"},
        )
    user["groups"].append(doc["group"])
    # Update the user
    results = await call(
        "user/update",
        skip_events=[Event.PERM],
        env=env,
        query=query,
        doc={"groups": user["groups"]},
    )
    # if update fails, return update results
    if results["status"] != 200:
        return results
    # Check if the updated User doc belongs to current session and update it
    if env["session"]["user"]["_id"] == user["_id"]:
        user_results = await call(
            "user/read_privileges",
            skip_events=[Event.PERM],
            env=env,
            query=[{"_id": user["_id"]}],
        )
        env["session"]["user"] = user_results["args"]["docs"][0]

    return results


async def _delete_group(env: "NawahEnv", query: "Query") -> "Results":
    # Confirm group is valid
    results = await call(
        "group/read",
        skip_events=[Event.PERM],
        env=env,
        query=[{"_id": query["group"][0]}],
    )
    if not results["args"]["count"]:
        raise FuncException(
            status=400, msg="Group is invalid.", args={"code": "INVALID_GROUP"}
        )
    # Get user details
    results = await call(
        "user/read", skip_events=[Event.PERM], env=env, query=[{"_id": query["_id"][0]}]
    )
    if not results["args"]["count"]:
        raise FuncException(
            status=400, msg="User is invalid.", args={"code": "INVALID_USER"}
        )
    user = results["args"]["docs"][0]
    # Confirm group was not added before
    if query["group"][0] not in user["groups"]:
        raise FuncException(
            status=400,
            msg="User is not a member of the group.",
            args={"code": "GROUP_NOT_ADDED"},
        )
    # Update the user
    results = await call(
        "user/update",
        skip_events=[Event.PERM],
        env=env,
        query=[{"_id": query["_id"][0]}],
        doc={"groups": {"$del_val": [query["group"][0]]}},
    )
    # if update fails, return update results
    if results["status"] != 200:
        return results
    # Check if the updated User doc belongs to current session and update it
    if env["session"]["user"]["_id"] == user["_id"]:
        user_results = await call(
            "user/read_privileges",
            skip_events=[Event.PERM],
            env=env,
            query=[{"_id": user["_id"]}],
        )
        env["session"]["user"] = user_results["args"]["docs"][0]

    return results



"""Provides 'create' Data Function"""

from typing import TYPE_CHECKING

from nawah.config import Config

if TYPE_CHECKING:
    from nawah.types import NawahDoc, NawahEnv, ResultsArgs


async def create(
    *,
    env: "NawahEnv",
    collection_name: str,
    doc: "NawahDoc",
) -> "ResultsArgs":
    """Creates 'doc' in 'collection' using connection 'conn' in 'env'"""

    collection = env["conn"][Config.data_name][collection_name]
    results = await collection.insert_one(doc)
    _id = results.inserted_id
    return {"count": 1, "docs": [{"_id": _id}]}



"""Provides types used in Nawah"""

from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    MutableMapping,
    MutableSequence,
    TypedDict,
    Union,
)

from nawah.enums import Event

if TYPE_CHECKING:
    from datetime import datetime

    from aiohttp.web import WebSocketResponse
    from motor.motor_asyncio import AsyncIOMotorClient

    from nawah.classes import Package


class Results(TypedDict):
    """Provides type-hint for return of Nawah Function"""

    status: int
    msg: str
    args: "ResultsArgs"


ResultsArgs = MutableMapping[str, Any]


class AppPackage(TypedDict):
    """Provides type-hint for item of 'Config.packages'"""

    package: "Package"
    modules: MutableSequence[str]


class AnalyticsEvents(TypedDict):
    """Provides type-hint for 'analytics_events' App Config Attr"""

    app_conn_verified: bool
    session_conn_auth: bool
    session_user_auth: bool
    session_conn_reauth: bool
    session_user_reauth: bool
    session_conn_deauth: bool
    session_user_deauth: bool


class IPQuota(TypedDict):
    """Provides type-hint for dict used to track IP user quota"""

    counter: int
    last_check: "datetime"


NawahEvents = MutableSequence[Event]


class NawahEnv(TypedDict, total=False):
    """Provides type-hint for 'Env' dict"""

    id: str
    init: bool
    conn: "AsyncIOMotorClient"
    REMOTE_ADDR: str
    HTTP_USER_AGENT: str
    HTTP_ORIGIN: str
    client_app: str
    session: "NawahSession"
    prev_session: MutableMapping
    prev_session_timeout: "datetime"
    last_call: "datetime"
    ws: "WebSocketResponse"
    quota: "NawahEnvQuota"
    args: MutableMapping[str, Any]


class NawahEnvQuota(TypedDict):
    """Provides type-hint for 'quota' of 'Env' dict"""

    counter: int
    last_check: "datetime"


class NawahSession(TypedDict):
    """Provieds type-hint for 'session' of 'Env' dict"""

    _id: str
    user: MutableMapping
    groups: MutableSequence[str]
    host_add: str
    user_agent: str
    expiry: str
    token_hash: str
    create_time: str


class NawahQuerySpecialGroup(TypedDict):
    """Provides type-hint for '$group' in 'NawahQuery'"""

    by: str
    count: int


class NawahQuerySpecialGeoNear(TypedDict):
    """Provides type-hint for '$geo_near' in 'NawahQuery'"""

    val: str
    attr: str
    dist: int


# Following TypedDict type can't be defined as class as keys include $
NawahQuerySpecial = TypedDict(
    "NawahQuerySpecial",
    {
        "$search": str,
        "$sort": MutableMapping[str, Literal[1, -1]],
        "$skip": int,
        "$limit": int,
        "$extn": Union[bool, MutableSequence[str]],
        "$attrs": MutableSequence[str],
        "$group": MutableSequence[NawahQuerySpecialGroup],
        "$geo_near": NawahQuerySpecialGeoNear,
    },
    total=False,
)

NawahQueryOperEq = TypedDict("NawahQueryOperEq", {"$eq": Any})
NawahQueryOperNe = TypedDict("NawahQueryOperNe", {"$ne": Any})
NawahQueryOperGt = TypedDict("NawahQueryOperGt", {"$gt": Union[int, float, str]})
NawahQueryOperGte = TypedDict("NawahQueryOperGte", {"$gte": Union[int, float, str]})
NawahQueryOperLt = TypedDict("NawahQueryOperLt", {"$lt": Union[int, float, str]})
NawahQueryOperLte = TypedDict("NawahQueryOperLte", {"$lte": Union[int, float, str]})
NawahQueryOperAll = TypedDict("NawahQueryOperAll", {"$all": list[Any]})
NawahQueryOperIn = TypedDict("NawahQueryOperIn", {"$in": list[Any]})
NawahQueryOperNin = TypedDict("NawahQueryOperNin", {"$nin": list[Any]})
NawahQueryOperRegex = TypedDict("NawahQueryOperRegex", {"$regex": str})
NawahQueryStep = dict[
    str,
    Union[
        NawahQueryOperEq,
        NawahQueryOperNe,
        NawahQueryOperGt,
        NawahQueryOperGte,
        NawahQueryOperLt,
        NawahQueryOperLte,
        NawahQueryOperAll,
        NawahQueryOperIn,
        NawahQueryOperNin,
        NawahQueryOperRegex,
    ],
]

NawahQueryOperOr = TypedDict("NawahQueryOperOr", {"$or": list[NawahQueryStep]})
NawahQueryOperAnd = TypedDict("NawahQueryOperAnd", {"$and": list[NawahQueryStep]})

NawahQueryIndex = TypedDict(
    "NawahQueryIndex", {"$index": dict[str, list[tuple[str, Any]]]}
)

NawahDoc = MutableMapping[str, Any]



import datetime
import logging
import re
from typing import TYPE_CHECKING

from nawah.config import Config
from nawah.enums import AttrType

if TYPE_CHECKING:
    from nawah.classes import Attr, Module

logger = logging.getLogger("nawah")


def generate_models_typescript():
    # Initialise _api_models Config Attr
    api_models = "// Nawah Models\n"
    # Add interface for DOC, LOCALE, LOCALES, FILE typing
    api_models += "interface Doc { _id: string; };\n"
    api_models += "export interface LOCALE { "
    for locale in Config.locales:
        api_models += locale
        if locale != Config.locale:
            api_models += "?"
        api_models += ": string; "
    api_models += "};\n"
    api_models += "export type LOCALES = '" + "' | '".join(Config.locales) + "';\n"
    api_models += "export type ID<T> = string & T;\n"
    api_models += "export interface FILE<T> { name: string; lastModified: number; type: T; size: number; content: string | boolean; };\n"
    # Iterate over packages in ascending order
    for package in sorted(Config.sys.packages):
        # Add package header
        api_models += f'\n// Package: {package.replace("modules.", "")}\n'
        if not Config.sys.packages[package]:
            api_models += "// No modules\n"
        # Iterate over package modules in ascending order
        for module in sorted(Config.sys.packages[package]):
            module_class = (
                str(Config.modules[module].__class__).split(".")[-1].split("'")[0]
            )
            # Add module header
            api_models += f"// Module: {module_class}\n"

            # Add module interface definition
            api_models += f"export interface {module_class} extends String, Doc {{\n"
            # Iterate over module attrs to add attrs types, defaults (if any)
            for attr in Config.modules[module].attrs.keys():
                attr_model = ""
                if Config.modules[module].attrs[attr].desc:
                    attr_model += f"\t// @property {{__TYPE__}} {Config.modules[module].attrs[attr].desc}\n"
                attr_model += f"\t{attr}__DEFAULT__: __TYPE__;\n"
                for default_attr in Config.modules[module].defaults.keys():
                    if default_attr == attr or default_attr.startswith(f"{attr}."):
                        # Attr is in defaults, indicate the same
                        attr_model = attr_model.replace("__DEFAULT__", "?")
                # Attempt to replace __DEFAULT__ with empty string if it still exists, effectively no default value
                attr_model = attr_model.replace("__DEFAULT__", "")

                # Add typing
                attr_model = attr_model.replace(
                    "__TYPE__",
                    _generate_model_typing(
                        module=Config.modules[module],
                        attr_name=attr,
                        attr_type=Config.modules[module].attrs[attr],
                    ),
                )

                api_models += attr_model

            # Add closing braces
            api_models += "};\n"
        api_models += "\n"
    import os

    models_file = os.path.join(
        os.path.abspath("."),
        "models",
        f'NAWAH_API_MODELS_{datetime.datetime.utcnow().strftime("%d-%b-%Y")}.ts',
    )
    with open(models_file, "w") as f:
        f.write(api_models)
        logger.info(f"API models generated and saved to: '{models_file}'. Exiting.")
        exit(0)


def _generate_model_typing(*, module: "Module", attr_name: str, attr_type: "Attr"):
    if attr_type.type == AttrType.ANY:
        return "any"

    if attr_type.type == AttrType.BOOL:
        return "boolean"

    if attr_type.type == AttrType.DATE:
        return "string"

    if attr_type.type == AttrType.DATETIME:
        return "string"

    if attr_type.type == AttrType.ATTR:
        return "{ type: string; args: { [key: string]: any; }; allow_none?: boolean; default: any; }"

    if attr_type.type == AttrType.KV_DICT:
        key_typing = _generate_model_typing(
            module=module, attr_name=attr_name, attr_type=attr_type.args["key"]
        )
        val_typing = _generate_model_typing(
            module=module, attr_name=attr_name, attr_type=attr_type.args["val"]
        )
        return f"{{ [key: {key_typing}]: {val_typing} }}"

    if attr_type.type == AttrType.TYPED_DICT:
        typing = "{ "
        for child_attr_type in attr_type.args["dict"].keys():
            typing += child_attr_type
            typing += ": "
            typing += _generate_model_typing(
                module=module,
                attr_name=attr_name,
                attr_type=attr_type.args["dict"][child_attr_type],
            )
            typing += "; "
        typing += "}"
        return typing

    if attr_type.type == AttrType.EMAIL:
        return "string"

    if attr_type.type == AttrType.FILE:
        types = "string"
        if attr_type.args["types"]:
            types = "'" + "' | '".join(attr_type.args["types"]) + "'"
            if "*" in types:
                types = "string"

        return f"FILE<{types}>"

    if attr_type.type == AttrType.FLOAT:
        return "number"

    if attr_type.type == AttrType.GEO_POINT:
        return "{ type: 'Point'; coordinates: [number, number]; }"

    if attr_type.type == AttrType.ID:
        for attr in module.extns.keys():
            # [TODO] Validate whether Extension Instructions as  breaks this code
            if attr.split(".")[0].split(":")[0] == attr_name:
                # [REF]: https://dev.to/rrampage/snake-case-to-camel-case-and-back-using-regular-expressions-and-python-m9j

                extn_module_class = re.sub(
                    r"(.*?)_([a-zA-Z])",
                    lambda match: match.group(1) + match.group(2).upper(),
                    module.extns[attr].module,
                )
                return f"ID<{extn_module_class[0].upper()}{extn_module_class[1:]}>"
        return "ID<string>"

    if attr_type.type == AttrType.INT:
        return "number"

    if attr_type.type == AttrType.IP:
        return "string"

    if attr_type.type == AttrType.LIST:
        list_typings = []
        for child_attr_type in attr_type.args["list"]:
            list_typings.append(
                _generate_model_typing(
                    module=module, attr_name=attr_name, attr_type=child_attr_type
                )
            )
        list_typing = " | ".join(list_typings)
        return f"Array<{list_typing}>"

    if attr_type.type == AttrType.LOCALE:
        return "LOCALE"

    if attr_type.type == AttrType.LOCALES:
        return "LOCALES"

    if attr_type.type == AttrType.PHONE:
        return "string"

    if attr_type.type == AttrType.STR:
        return "string"

    if attr_type.type == AttrType.TIME:
        return "string"

    if attr_type.type == AttrType.URI_WEB:
        return "string"

    if attr_type.type == AttrType.LITERAL:
        return "'" + "' | '".join(attr_type.args["literal"]) + "'"

    if attr_type.type == AttrType.UNION:
        return " | ".join(
            [
                _generate_model_typing(
                    module=module, attr_name=attr_name, attr_type=child_attr_type
                )
                for child_attr_type in attr_type.args["union"]
            ]
        )

    if attr_type.type == AttrType.TYPE:
        return "any"



"""Provides 'validate_doc', 'validate_attr' Utilities"""

import copy
import datetime
import inspect
import logging
import re
from math import inf
from typing import TYPE_CHECKING, Any, Literal, MutableMapping, Union, cast

from bson import ObjectId

from nawah.classes import Attr
from nawah.config import Config
from nawah.enums import AttrType, LocaleStrategy
from nawah.exceptions import InvalidAttrException, MissingAttrException

if TYPE_CHECKING:
    from nawah.types import NawahDoc

logger = logging.getLogger("nawah")

file_typed_dict_attr = Attr.TYPED_DICT(
    dict={
        "name": Attr.STR(),
        "lastModified": Attr.INT(),
        "type": Attr.STR(),
        "size": Attr.INT(),
        "ref": Attr.ID(),
    }
)


def validate_doc(
    *,
    mode: Literal["create", "create_draft", "update"],
    doc: "NawahDoc",
    attrs: MutableMapping[str, "Attr"],
):
    """Validates 'doc' value against dict of Attr Types dict 'attrs'"""

    attrs_map = {attr.split(".")[0]: attr for attr in doc.keys()}

    for attr in attrs:
        if attr not in attrs_map:
            if mode == "create":
                doc[attr] = None
            else:
                continue
        elif mode != "create":
            attr = attrs_map[attr]

        if mode != "create" and "." in attr:
            doc[attr] = _validate_dot_notated(
                attr=attr,
                doc=doc,
                attrs=attrs,
            )
        else:
            try:
                doc[attr] = validate_attr(
                    attr_name=attr,
                    attr_type=attrs[attr],
                    attr_val=doc[attr],
                    mode=mode,
                    doc=doc,
                )
            except Exception as e:
                if doc[attr] is None:
                    raise MissingAttrException(attr_name=attr) from e

                raise InvalidAttrException(
                    attr_name=attr, attr_type=attrs[attr], val_type=type(doc[attr])
                ) from e


def _validate_dot_notated(
    attr: str,
    doc: "NawahDoc",
    attrs: MutableMapping[str, "Attr"],
):
    """Validates value of dot-notated 'attr' against Attr Type"""

    attr_path = attr.split(".")
    attr_path_len = len(attr_path)
    attr_type: Union[MutableMapping[str, "Attr"], "Attr"] = attrs

    try:
        for i in range(attr_path_len):
            # Iterate over attr_path to reach last valid Attr Type
            if not isinstance(attr_type, Attr):
                attr_type = cast(MutableMapping[str, "Attr"], attr_type)

                attr_type = attr_type[attr_path[i]]
            elif isinstance(attr_type, Attr):
                attr_type = cast("Attr", attr_type)

                if attr_type.type == AttrType.ANY:
                    return doc[attr]

                if attr_type.type == AttrType.LOCALE:
                    if attr_path[i] not in Config.locales:
                        raise Exception()
                    attr_type = Attr.STR()
                elif attr_type.type == AttrType.TYPED_DICT:
                    attr_type = attr_type.args["dict"][attr_path[i]]
                elif attr_type.type == AttrType.KV_DICT:
                    attr_type = attr_type.args["val"]
                # However, if list or union, start a new _validate_dot_notated call as it is required to check all the provided types
                elif attr_type.type in [AttrType.LIST, AttrType.UNION]:
                    if attr_type.type == AttrType.LIST:
                        attr_type_iter = attr_type.args["list"]
                    else:
                        attr_type_iter = attr_type.args["union"]
                    for child_attr_type in attr_type_iter:
                        attr_val = _validate_dot_notated(
                            attr=".".join(attr_path[i:]),
                            doc={".".join(attr_path[i:]): doc[attr]},
                            attrs={attr_path[i]: child_attr_type},
                        )
                        if attr_val is not None:
                            return attr_val
                    raise Exception()
                else:
                    raise Exception()
            else:
                raise Exception()

        # Validate val against final Attr Type
        # mode is statically set to update as dot-notation attrs are only allowed in update calls
        attr_val = validate_attr(
            mode="update",
            attr_name=attr,
            attr_type=cast("Attr", attr_type),
            attr_val=doc[attr],
            doc=doc,
        )
        return attr_val
    except Exception as e:
        raise InvalidAttrException(
            attr_name=attr, attr_type=attrs[attr_path[0]], val_type=type(doc[attr])
        ) from e


def validate_attr(
    *,
    mode: Literal["create", "create_draft", "update", "deep"],
    attr_name: str,
    attr_type: "Attr",
    attr_val: Any,
    doc: "NawahDoc" = None,
):
    """Validates value 'attr_val' against Attr Type 'attr_type'. Returns processed valid
    value, if valid, otherwise raises 'InvalidAttrException'"""

    # Basic checks for None-value
    if attr_val is None:
        # For mode==update, having None-value is accepted
        if mode == "update":
            return None
        # Otherwise check for default
        if attr_type.default:
            # If default.value is callable, call with doc to generate value
            if callable(attr_type.default.value):
                return attr_type.default.value(doc)
            # Else, return copy from default.value (to avoid passing objects)
            return copy.deepcopy(attr_type.default.value)

    attr_oper: Literal[
        None, "$add", "$multiply", "$append", "$set_index", "$del_val", "$del_index"
    ] = None
    attr_oper_args = {}
    if mode == "update" and isinstance(attr_val, dict):
        if "$add" in attr_val.keys():
            attr_oper = "$add"
            if "$field" in attr_val.keys() and attr_val["$field"]:
                attr_oper_args["$field"] = attr_val["$field"]
            else:
                attr_oper_args["$field"] = None
            attr_val = attr_val["$add"]
        elif "$multiply" in attr_val.keys():
            attr_oper = "$multiply"
            if "$field" in attr_val.keys() and attr_val["$field"]:
                attr_oper_args["$field"] = attr_val["$field"]
            else:
                attr_oper_args["$field"] = None
            attr_val = attr_val["$multiply"]
        elif "$append" in attr_val.keys():
            attr_oper = "$append"
            if "$unique" in attr_val.keys() and attr_val["$unique"] == True:
                attr_oper_args["$unique"] = True
            else:
                attr_oper_args["$unique"] = False
            attr_val = [attr_val["$append"]]
        elif "$set_index" in attr_val.keys():
            attr_oper = "$set_index"
            attr_oper_args["$index"] = attr_val["$index"]
            attr_val = [attr_val["$set_index"]]
        elif "$del_val" in attr_val.keys():
            attr_oper = "$del_val"
            attr_val = attr_val["$del_val"]
            if attr_type.type != "LIST" or type(attr_val) != list:
                raise InvalidAttrException(
                    attr_name=attr_name, attr_type=attr_type, val_type=type(attr_val)
                )
            return return_valid_attr(
                attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
            )
        elif "$del_index" in attr_val.keys():
            attr_oper = "$del_index"
            attr_oper_args["$index"] = attr_val["$del_index"]
            attr_val = attr_val["$del_index"]
            if (attr_type.type == AttrType.LIST and type(attr_val) == int) or (
                attr_type.type == AttrType.KV_DICT and type(attr_val) == str
            ):
                return return_valid_attr(
                    attr_val=attr_val,
                    attr_oper=attr_oper,
                    attr_oper_args=attr_oper_args,
                )
            else:
                raise InvalidAttrException(
                    attr_name=attr_name, attr_type=attr_type, val_type=type(attr_val)
                )

    # Deepcopy attr_val to eliminate changes in in original object
    attr_val = copy.deepcopy(attr_val)

    if attr_type.type == AttrType.ANY:
        if attr_val is not None:
            return return_valid_attr(
                attr_val=attr_val,
                attr_oper=attr_oper,
                attr_oper_args=attr_oper_args,
            )

    elif attr_type.type == AttrType.BOOL:
        if type(attr_val) == bool:
            return return_valid_attr(
                attr_val=attr_val,
                attr_oper=attr_oper,
                attr_oper_args=attr_oper_args,
            )

    elif attr_type.type == AttrType.DATE:
        if isinstance(attr_val, str) and re.match(
            r"^[0-9]{4}-[0-9]{2}-[0-9]{2}$", attr_val
        ):
            if attr_type.args["ranges"]:
                for date_range in attr_type.args["ranges"]:
                    date_range = copy.deepcopy(date_range)
                    for i in [0, 1]:
                        if date_range[i][0] in ["+", "-"]:
                            date_range_delta = {}
                            if date_range[i][-1] == "d":
                                date_range_delta = {"days": int(date_range[i][:-1])}
                            elif date_range[i][-1] == "w":
                                date_range_delta = {"weeks": int(date_range[i][:-1])}
                            date_range[i] = (
                                (
                                    datetime.datetime.utcnow()
                                    + datetime.timedelta(**date_range_delta)
                                )
                                .isoformat()
                                .split("T")[0]
                            )
                    if attr_val >= date_range[0] and attr_val < date_range[1]:
                        return return_valid_attr(
                            attr_val=attr_val,
                            attr_oper=attr_oper,
                            attr_oper_args=attr_oper_args,
                        )
            else:
                return return_valid_attr(
                    attr_val=attr_val,
                    attr_oper=attr_oper,
                    attr_oper_args=attr_oper_args,
                )

    elif attr_type.type == AttrType.DATETIME:
        if isinstance(attr_val, str) and re.match(
            r"^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}(:[0-9]{2}(\.[0-9]{6})?)?$",
            attr_val,
        ):
            if attr_type.args["ranges"]:
                for datetime_range in attr_type.args["ranges"]:
                    datetime_range = copy.deepcopy(datetime_range)
                    for i in [0, 1]:
                        if datetime_range[i][0] in ["+", "-"]:
                            datetime_range_delta = {}
                            if datetime_range[i][-1] == "d":
                                datetime_range_delta = {
                                    "days": int(datetime_range[i][:-1])
                                }
                            elif datetime_range[i][-1] == "s":
                                datetime_range_delta = {
                                    "seconds": int(datetime_range[i][:-1])
                                }
                            elif datetime_range[i][-1] == "m":
                                datetime_range_delta = {
                                    "minutes": int(datetime_range[i][:-1])
                                }
                            elif datetime_range[i][-1] == "h":
                                datetime_range_delta = {
                                    "hours": int(datetime_range[i][:-1])
                                }
                            elif datetime_range[i][-1] == "w":
                                datetime_range_delta = {
                                    "weeks": int(datetime_range[i][:-1])
                                }
                            datetime_range[i] = (
                                datetime.datetime.utcnow()
                                + datetime.timedelta(**datetime_range_delta)
                            ).isoformat()
                    if attr_val >= datetime_range[0] and attr_val < datetime_range[1]:
                        return return_valid_attr(
                            attr_val=attr_val,
                            attr_oper=attr_oper,
                            attr_oper_args=attr_oper_args,
                        )
            else:
                return return_valid_attr(
                    attr_val=attr_val,
                    attr_oper=attr_oper,
                    attr_oper_args=attr_oper_args,
                )

    elif attr_type.type == AttrType.ATTR:
        if isinstance(attr_val, dict):
            # [TODO] Clean-up to import this on root level
            from .._attr_type import decode_attr_type

            decode_attr_type(encoded_attr_type=attr_val)
            return return_valid_attr(
                attr_val=attr_val,
                attr_oper=attr_oper,
                attr_oper_args=attr_oper_args,
            )

    elif attr_type.type == AttrType.KV_DICT:
        if type(attr_val) == dict:
            if attr_type.args["len_range"]:
                if not (
                    attr_type.args["len_range"][0]
                    <= len(attr_val.keys())
                    < attr_type.args["len_range"][1]
                ):
                    raise InvalidAttrException(
                        attr_name=attr_name,
                        attr_type=attr_type,
                        val_type=type(attr_val),
                    )
            if attr_type.args["req"]:
                for req_key in attr_type.args["req"]:
                    if req_key not in attr_val.keys():
                        raise InvalidAttrException(
                            attr_name=attr_name,
                            attr_type=attr_type,
                            val_type=type(attr_val),
                        )
            shadow_attr_val = {}
            try:
                for child_attr_val in attr_val.keys():
                    shadow_attr_val[
                        validate_attr(
                            mode="deep",
                            attr_name=f"{attr_name}.{child_attr_val}",
                            attr_type=attr_type.args["key"],
                            attr_val=child_attr_val,
                            doc=doc,
                        )
                    ] = validate_attr(
                        mode="deep",
                        attr_name=f"{attr_name}.{child_attr_val}",
                        attr_type=attr_type.args["val"],
                        attr_val=attr_val[child_attr_val],
                        doc=doc,
                    )
                return return_valid_attr(
                    attr_val=shadow_attr_val,
                    attr_oper=attr_oper,
                    attr_oper_args=attr_oper_args,
                )
            except InvalidAttrException:
                pass

    elif attr_type.type == AttrType.TYPED_DICT:
        if type(attr_val) == dict:
            try:
                for child_attr_type in attr_type.args["dict"].keys():
                    if child_attr_type not in attr_val.keys():
                        attr_val[child_attr_type] = None
                    attr_val[child_attr_type] = validate_attr(
                        mode="deep",
                        attr_name=f"{attr_name}.{child_attr_type}",
                        attr_type=attr_type.args["dict"][child_attr_type],
                        attr_val=attr_val[child_attr_type],
                        doc=doc,
                    )
                # Match keys _after_ checking child attrs in order to allow _validate_default to run on all child attrs
                if set(attr_val.keys()) != set(attr_type.args["dict"].keys()):
                    raise InvalidAttrException(
                        attr_name=attr_name,
                        attr_type=attr_type,
                        val_type=type(attr_val),
                    )

                return return_valid_attr(
                    attr_val=attr_val,
                    attr_oper=attr_oper,
                    attr_oper_args=attr_oper_args,
                )
            except InvalidAttrException:
                pass

    elif attr_type.type == AttrType.EMAIL:
        if type(attr_val) == str and re.match(r"^[^@]+@[^@]+\.[^@]+$", attr_val):
            if attr_type.args["allowed_domains"]:
                for domain in attr_type.args["allowed_domains"]:
                    if attr_type.args["strict_matching"]:
                        domain = f"@{domain}"
                    if attr_val.endswith(domain):
                        return return_valid_attr(
                            attr_val=attr_val,
                            attr_oper=attr_oper,
                            attr_oper_args=attr_oper_args,
                        )
            elif attr_type.args["disallowed_domains"]:
                for domain in attr_type.args["disallowed_domains"]:
                    if attr_type.args["strict_matching"]:
                        domain = f"@{domain}"
                    if attr_val.endswith(domain):
                        break
                else:
                    return return_valid_attr(
                        attr_val=attr_val,
                        attr_oper=attr_oper,
                        attr_oper_args=attr_oper_args,
                    )
            else:
                return return_valid_attr(
                    attr_val=attr_val,
                    attr_oper=attr_oper,
                    attr_oper_args=attr_oper_args,
                )

    elif attr_type.type == AttrType.FILE:
        try:
            if isinstance(attr_val, list) and len(attr_val):
                attr_val = validate_attr(
                    mode="deep",
                    attr_name=attr_name,
                    attr_type=attr_type,
                    attr_val=attr_val[0],
                    doc=doc,
                )

            validate_attr(
                mode="create",
                attr_name=attr_name,
                attr_type=file_typed_dict_attr,
                attr_val=attr_val,
            )

            return return_valid_attr(
                attr_val=attr_val,
                attr_oper=attr_oper,
                attr_oper_args=attr_oper_args,
            )

        except InvalidAttrException:
            pass

    elif attr_type.type == AttrType.BYTES:
        if isinstance(attr_val, bytes):
            return return_valid_attr(
                attr_val=attr_val,
                attr_oper=attr_oper,
                attr_oper_args=attr_oper_args,
            )

    elif attr_type.type == AttrType.FLOAT:
        if isinstance(attr_val, str) and re.match(r"^[0-9]+(\.[0-9]+)?$", attr_val):
            attr_val = float(attr_val)
        elif isinstance(attr_val, int):
            attr_val = float(attr_val)

        if isinstance(attr_val, float):
            if attr_type.args["ranges"]:
                for _range in attr_type.args["ranges"]:
                    if _range[0] <= attr_val < _range[1]:
                        return return_valid_attr(
                            attr_val=attr_val,
                            attr_oper=attr_oper,
                            attr_oper_args=attr_oper_args,
                        )
            else:
                return return_valid_attr(
                    attr_val=attr_val,
                    attr_oper=attr_oper,
                    attr_oper_args=attr_oper_args,
                )

    elif attr_type.type == AttrType.GEO_POINT:
        if (
            isinstance(attr_val, dict)
            and set(attr_val.keys()) == {"type", "coordinates"}
            and attr_val["type"] in ["Point"]
            and isinstance(attr_val["coordinates"], list)
            and len(attr_val["coordinates"]) == 2
            and type(attr_val["coordinates"][0]) in [int, float]
            and type(attr_val["coordinates"][1]) in [int, float]
        ):
            return return_valid_attr(
                attr_val=attr_val,
                attr_oper=attr_oper,
                attr_oper_args=attr_oper_args,
            )

    elif attr_type.type == AttrType.ID:
        if isinstance(attr_val, dict) and "_id" in attr_val:
            return return_valid_attr(
                attr_val=str(attr_val["_id"]),
                attr_oper=attr_oper,
                attr_oper_args=attr_oper_args,
            )

        if isinstance(attr_val, ObjectId):
            return return_valid_attr(
                attr_val=str(attr_val),
                attr_oper=attr_oper,
                attr_oper_args=attr_oper_args,
            )

        if isinstance(attr_val, str):
            try:
                if not re.match(r"^[a-f0-9]{24}$", attr_val):
                    raise InvalidAttrException(
                        attr_name=attr_name, attr_type=attr_type, val_type=str
                    )

                return return_valid_attr(
                    attr_val=attr_val,
                    attr_oper=attr_oper,
                    attr_oper_args=attr_oper_args,
                )
            except InvalidAttrException:
                pass

    elif attr_type.type == AttrType.INT:
        if isinstance(attr_val, str) and re.match(r"^[0-9]+$", attr_val):
            attr_val = int(attr_val)

        if isinstance(attr_val, int):
            if attr_type.args["ranges"]:
                # [TODO] Implement inf range, ref: https://stackoverflow.com/questions/9884213/looping-from-1-to-infinity-in-python
                for _range in attr_type.args["ranges"]:
                    if attr_val in range(*_range):
                        return return_valid_attr(
                            attr_val=attr_val,
                            attr_oper=attr_oper,
                            attr_oper_args=attr_oper_args,
                        )
            else:
                return return_valid_attr(
                    attr_val=attr_val,
                    attr_oper=attr_oper,
                    attr_oper_args=attr_oper_args,
                )

    elif attr_type.type == AttrType.IP:
        if isinstance(attr_val, str) and re.match(
            r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$",
            attr_val,
        ):
            return return_valid_attr(
                attr_val=attr_val,
                attr_oper=attr_oper,
                attr_oper_args=attr_oper_args,
            )

    elif attr_type.type == AttrType.LIST:
        if isinstance(attr_val, list):
            try:
                if attr_type.args["len_range"]:
                    if not (
                        attr_type.args["len_range"][0]
                        <= len(attr_val)
                        < attr_type.args["len_range"][1]
                    ):
                        raise InvalidAttrException(
                            attr_name=attr_name,
                            attr_type=attr_type,
                            val_type=type(attr_val),
                        )
                for i, child_attr_val in enumerate(attr_val):
                    child_attr_check = False
                    for child_attr_type in attr_type.args["list"]:
                        try:
                            attr_val[i] = validate_attr(
                                mode="deep",
                                attr_name=attr_name,
                                attr_type=child_attr_type,
                                attr_val=child_attr_val,
                                doc=doc,
                            )
                            child_attr_check = True
                            break
                        except InvalidAttrException:
                            pass

                    if not child_attr_check:
                        raise InvalidAttrException(
                            attr_name=attr_name,
                            attr_type=attr_type,
                            val_type=type(attr_val),
                        )
                return return_valid_attr(
                    attr_val=attr_val,
                    attr_oper=attr_oper,
                    attr_oper_args=attr_oper_args,
                )
            except InvalidAttrException:
                pass

    elif attr_type.type == AttrType.LOCALE:
        try:
            attr_val = validate_attr(
                mode="deep",
                attr_name=attr_name,
                attr_type=Attr.KV_DICT(
                    key=Attr.LITERAL(literal=copy.copy(Config.locales)),
                    val=Attr.STR(),
                    len_range=[1, inf],
                    req=[Config.locale],
                ),
                attr_val=attr_val,
                doc=doc,
            )
            if Config.locale_strategy == LocaleStrategy.NONE_VALUE:
                attr_val = {
                    locale: attr_val[locale] if locale in attr_val.keys() else None
                    for locale in Config.locales
                }
            elif callable(Config.locale_strategy):
                attr_val = {
                    locale: attr_val[locale]
                    if locale in attr_val.keys()
                    else Config.locale_strategy(attr_val=attr_val, locale=locale)
                    for locale in Config.locales
                }
            else:
                attr_val = {
                    locale: attr_val[locale]
                    if locale in attr_val.keys()
                    else attr_val[Config.locale]
                    for locale in Config.locales
                }
            return return_valid_attr(
                attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
            )
        except InvalidAttrException:
            pass

    elif attr_type.type == AttrType.LOCALES:
        if attr_val in Config.locales:
            return return_valid_attr(
                attr_val=attr_val,
                attr_oper=attr_oper,
                attr_oper_args=attr_oper_args,
            )

    elif attr_type.type == AttrType.PHONE:
        if isinstance(attr_val, str):
            if attr_type.args["codes"]:
                for phone_code in attr_type.args["codes"]:
                    if re.match(rf"^\+{phone_code}[0-9]+$", attr_val):
                        return return_valid_attr(
                            attr_val=attr_val,
                            attr_oper=attr_oper,
                            attr_oper_args=attr_oper_args,
                        )
            else:
                if re.match(r"^\+[0-9]+$", attr_val):
                    return return_valid_attr(
                        attr_val=attr_val,
                        attr_oper=attr_oper,
                        attr_oper_args=attr_oper_args,
                    )

    elif attr_type.type == AttrType.STR:
        if isinstance(attr_val, str):
            if attr_type.args["pattern"]:
                if re.match(f'^{attr_type.args["pattern"]}$', attr_val):
                    return return_valid_attr(
                        attr_val=attr_val,
                        attr_oper=attr_oper,
                        attr_oper_args=attr_oper_args,
                    )
            else:
                return return_valid_attr(
                    attr_val=attr_val,
                    attr_oper=attr_oper,
                    attr_oper_args=attr_oper_args,
                )

    elif attr_type.type == AttrType.TIME:
        if isinstance(attr_val, str) and re.match(
            r"^[0-9]{2}:[0-9]{2}(:[0-9]{2}(\.[0-9]{6})?)?$", attr_val
        ):
            if attr_type.args["ranges"]:
                for time_range in attr_type.args["ranges"]:
                    time_range = copy.deepcopy(time_range)
                    for i in [0, 1]:
                        if time_range[i][0] in ["+", "-"]:
                            time_range_delta = {}
                            if time_range[i][-1] == "s":
                                time_range_delta = {"seconds": int(time_range[i][:-1])}
                            elif time_range[i][-1] == "m":
                                time_range_delta = {"minutes": int(time_range[i][:-1])}
                            elif time_range[i][-1] == "h":
                                time_range_delta = {"hours": int(time_range[i][:-1])}
                            time_range[i] = (
                                (
                                    datetime.datetime.utcnow()
                                    + datetime.timedelta(**time_range_delta)
                                )
                                .isoformat()
                                .split("T")[1]
                            )
                    if time_range[1] > attr_val >= time_range[0]:
                        return return_valid_attr(
                            attr_val=attr_val,
                            attr_oper=attr_oper,
                            attr_oper_args=attr_oper_args,
                        )
            else:
                return return_valid_attr(
                    attr_val=attr_val,
                    attr_oper=attr_oper,
                    attr_oper_args=attr_oper_args,
                )

    elif attr_type.type == AttrType.URI_WEB:
        if isinstance(attr_val, str) and re.match(
            r"^https?:\/\/(?:[\w\-\_]+\.)(?:\.?[\w]{2,})+([\?\/].*)?$", attr_val
        ):
            if attr_type.args["allowed_domains"]:
                attr_val_domain = attr_val.split("/")[2]
                for domain in attr_type.args["allowed_domains"]:
                    if attr_type.args["strict_matching"] and attr_val_domain == domain:
                        return return_valid_attr(
                            attr_val=attr_val,
                            attr_oper=attr_oper,
                            attr_oper_args=attr_oper_args,
                        )

                    if not attr_type.args[
                        "strict_matching"
                    ] and attr_val_domain.endswith(domain):
                        return return_valid_attr(
                            attr_val=attr_val,
                            attr_oper=attr_oper,
                            attr_oper_args=attr_oper_args,
                        )
            elif attr_type.args["disallowed_domains"]:
                attr_val_domain = attr_val.split("/")[2]
                for domain in attr_type.args["disallowed_domains"]:
                    if attr_type.args["strict_matching"] and attr_val_domain == domain:
                        break

                    if not attr_type.args[
                        "strict_matching"
                    ] and attr_val_domain.endswith(domain):
                        break
                else:
                    return return_valid_attr(
                        attr_val=attr_val,
                        attr_oper=attr_oper,
                        attr_oper_args=attr_oper_args,
                    )
            else:
                return return_valid_attr(
                    attr_val=attr_val,
                    attr_oper=attr_oper,
                    attr_oper_args=attr_oper_args,
                )

    elif attr_type.type == AttrType.LITERAL:
        if attr_val in attr_type.args["literal"]:
            return return_valid_attr(
                attr_val=attr_val,
                attr_oper=attr_oper,
                attr_oper_args=attr_oper_args,
            )

    elif attr_type.type == AttrType.UNION:
        child_attr_check = False
        for child_attr in attr_type.args["union"]:
            try:
                attr_val = validate_attr(
                    mode="deep",
                    attr_name=attr_name,
                    attr_type=child_attr,
                    attr_val=attr_val,
                    doc=doc,
                )
                child_attr_check = True
            except InvalidAttrException:
                continue

            if child_attr_check:
                return return_valid_attr(
                    attr_val=attr_val,
                    attr_oper=attr_oper,
                    attr_oper_args=attr_oper_args,
                )

    elif attr_type.type == AttrType.TYPE:
        try:
            func_params = {
                "mode": mode,
                "attr_name": attr_name,
                "attr_type": attr_type,
                "attr_val": attr_val,
            }
            attr_val = attr_type.args["func"](
                **{
                    param: func_params[param]
                    for param in inspect.signature(attr_type.args["func"]).parameters
                }
            )
            return return_valid_attr(
                attr_val=attr_val,
                attr_oper=attr_oper,
                attr_oper_args=attr_oper_args,
            )
        except InvalidAttrException:
            pass

    # If no case is matched, check mode, default
    # For mode==update, having None-value is accepted
    if mode == "update":
        return None
    # Otherwise check for default
    if attr_type.default:
        # If default.value is callable, call with doc to generate value
        if callable(attr_type.default.value):
            return attr_type.default.value(doc)
        # Else, return copy from default.value (to avoid passing objects)
        return copy.deepcopy(attr_type.default.value)

    # If no case is matched, raise
    raise InvalidAttrException(
        attr_name=attr_name, attr_type=attr_type, val_type=type(attr_val)
    )


def return_valid_attr(
    *,
    attr_val: Any,
    attr_oper: Literal[
        None, "$add", "$multiply", "$append", "$set_index", "$del_val", "$del_index"
    ],
    attr_oper_args: MutableMapping[str, Any],
) -> Any:
    if not attr_oper:
        return attr_val

    if attr_oper in ["$add", "$multiply"]:
        return {attr_oper: attr_val, "$field": attr_oper_args["$field"]}

    if attr_oper == "$del_val":
        return {attr_oper: attr_val}

    if attr_oper == "$append":
        return {"$append": attr_val[0], "$unique": attr_oper_args["$unique"]}

    if attr_oper == "$set_index":
        return {"$set_index": attr_val[0], "$index": attr_oper_args["$index"]}

    if attr_oper == "$del_index":
        return {"$del_index": attr_oper_args["$index"]}



"""Provides utilities required for configuring modules, packages and app"""

import asyncio
import copy
import datetime
import logging
import os
import re
from dataclasses import fields
from typing import TYPE_CHECKING, Any, MutableMapping, MutableSequence, cast

import sentry_sdk
from bson import ObjectId
from croniter import croniter
from passlib.hash import pbkdf2_sha512
from redis import asyncio as aioredis  # type: ignore

from nawah.classes import (
    App,
    Attr,
    Counter,
    Default,
    Diff,
    Extn,
    Func,
    Perm,
    SysDoc,
    UserAttr,
    Var,
)
from nawah.config import Config
from nawah.enums import AttrType, Event
from nawah.exceptions import (
    ConfigException,
    InvalidAttrTypeArgException,
    InvalidAttrTypeException,
    InvalidVarException,
)

from ._attr_type import extract_attr, generate_attr_val
from ._call import call
from ._val import deep_update, var_value
from ._validate import validate_attr, validate_type

if TYPE_CHECKING:
    from nawah.classes import Env, Module, Package
    from nawah.types import NawahDoc

logger = logging.getLogger("nawah")


def process_env_vars(config: "Env", /):
    """Checks all Var-prune Config Attrs of 'config' and updates value"""

    var_args: MutableMapping[str, MutableMapping[str, Any]] = {
        "debug": {"var": config.debug, "type": bool, "default": False},
        "port": {"var": config.port, "type": int, "default": 8081},
        "conn_timeout": {"var": config.conn_timeout, "type": int},
        "quota_anon_min": {"var": config.quota_anon_min, "type": int},
        "quota_auth_min": {"var": config.quota_auth_min, "type": int},
        "quota_ip_min": {"var": config.quota_ip_min, "type": int},
        "file_upload_limit": {
            "var": config.file_upload_limit,
            "type": int,
        },
        "file_upload_timeout": {
            "var": config.file_upload_timeout,
            "type": int,
        },
        "data_server": {"var": config.data_server, "type": str},
        "data_name": {"var": config.data_name, "type": str},
        "data_ssl": {"var": config.data_ssl, "type": bool},
        "data_disk_use": {"var": config.data_disk_use, "type": bool},
        "cache_server": {"var": config.cache_server, "type": str},
        "cache_db": {
            "var": config.cache_db,
            "type": int,
            "default": 0,
        },
        "cache_username": {"var": config.cache_username, "type": str},
        "cache_password": {"var": config.cache_password, "type": str},
        "cache_expiry": {"var": config.cache_expiry, "type": int},
        "error_reporting_server": {
            "var": config.error_reporting_server,
            "type": str,
        },
    }

    for arg_name, var_arg in var_args.items():
        if not isinstance(var_arg["var"], Var):
            continue
        try:
            setattr(
                config, arg_name, var_value(var_arg["var"], return_type=var_arg["type"])
            )
        except InvalidVarException as e:
            if "default" not in var_arg:
                raise e
            setattr(config, arg_name, var_arg["default"])


def _pre_setup_app(app_config: "App", /):
    # Check envs, env App Config Attr
    if app_config.envs and app_config.env:
        if isinstance(app_config.env, Var):
            logger.debug(
                "App '{name}' defines 'env' App Config Attr as 'Var' object. Getting its "
                "value.."
            )
            app_config.env = var_value(app_config.env)

        if app_config.env not in app_config.envs:
            raise ConfigException(
                "Value for 'env' App Config Attr is '{app_config.env}', but"
                "no such Env is defined"
            )

        # Update runtime config with values from Env object
        app_config.env = cast(str, app_config.env)
        env = app_config.envs[app_config.env]
        for env_field in fields(env):
            if (env_attr_val := getattr(env, env_field.name)) is not None:
                # Specifically for l10n Config Attr, always update base value, and don't overwrite it
                if env_field.name == "l10n":
                    # For a possibility app_config.l10n is None, set as L10N
                    if app_config.l10n is None:
                        app_config.l10n = {}
                    app_config.l10n.update(env_attr_val)
                else:
                    setattr(app_config, env_field.name, env_attr_val)

    if app_config.debug is True:
        logger.setLevel(logging.DEBUG)
        logger.debug("Set logging level to DEBUG (Config.debug==True)")

    if isinstance(app_config.debug, Var):
        try:
            if var_value(app_config.debug, return_type=lambda _: bool(int(_))):
                logger.setLevel(logging.DEBUG)
                logger.debug("Set logging level to DEBUG (Config.debug.var)")
                app_config.debug = True
        except InvalidVarException:
            logger.warning(
                "No value for 'Var' object for 'debug' App Config Attr. Logging "
                "remains at level ERROR"
            )

    logger.debug("Getting value of 'Var' object of 'admin_password' App Config Attr..")
    if isinstance(app_config.admin_password, Var):
        try:
            admin_password_val = var_value(app_config.admin_password)
            if not re.match(
                r"^(?=.*[a-z])(?=.*[A-Z])(?=.*[0-9])(?=.{8,})", admin_password_val
            ):
                raise Exception(
                    f"Value '{admin_password_val}' of 'admin_password' App Config Attr doesn't "
                    "match accepted password format"
                )
            app_config.admin_password = admin_password_val
            Config.sys.admin_check = True
        except InvalidVarException:
            logger.warning(
                "No value for 'Var' object for 'admin_password' App Config Attr. 'ADMIN' doc "
                "will not be checked or updated"
            )
            Config.sys.admin_check = False

    if isinstance(app_config.anon_token, Var):
        logger.debug(
            "App '{name}' defines 'anon_token' App Config Attr as 'Var' object. Getting its "
            "value.."
        )
        app_config.anon_token = var_value(app_config.anon_token)

    process_env_vars(app_config)


def setup_app(app_config: "App", /, *, process_vars: bool = False):
    # pylint: disable=import-outside-toplevel

    from nawah.packages.core import core

    app_config_version = cast(str, app_config.version)

    if not re.match(r"^[0-9]+\.[0-9]+\.[0-9]+([ab][0-9]+)?$", app_config_version):
        raise ConfigException(
            f"App '{app_config.name}' defines invalid 'version' '{app_config.version}'"
        )

    if process_vars:
        _pre_setup_app(app_config)

    configs: MutableSequence["Package"] = [
        core,
        *(app_config.packages or []),
        app_config,
    ]

    for config in configs:
        for config_field in fields(config):
            attr_name = config_field.name
            attr_val = getattr(config, attr_name, None)

            if attr_val is None:
                continue

            if attr_name in ["envs", "api_level", "version", "modules", "packages"]:
                continue

            # For user_attrs, process separatly to update user module
            if attr_name == "user_attrs":
                user_module = Config.modules["user"]
                for user_attr_name, user_attr in cast(
                    MutableMapping[str, "UserAttr"], config.user_attrs
                ).items():
                    if not isinstance(user_attr, UserAttr):
                        raise ConfigException(
                            f"Invalid type for attr '{user_attr_name}' defined for package "
                            f"'{config.name}' for 'user_attrs' Config Attr"
                        )
                    if user_attr_name in user_module.attrs:
                        raise ConfigException(
                            f"Package '{config.name}' defines 'user_attrs' Config Attr with "
                            f"conflicting attr '{user_attr_name}'"
                        )
                    if not isinstance(user_attr.type, Attr):
                        raise ConfigException(
                            f"Invalid type for attr type '{user_attr_name}' defined for package "
                            f"'{config.name}' for 'user_attrs' Config Attr"
                        )

                    Config.user_attrs[user_attr_name] = user_attr
                    user_module.attrs[user_attr_name] = user_attr.type

                    if user_attr.sanitise:
                        Config.user_attrs_sanitise.append(user_attr_name)

                    if user_attr.auth_attr:
                        Config.auth_attrs.append(user_attr_name)
                        Config.user_attrs_sanitise.append(f"{user_attr_name}_hash")
                        user_module.unique_attrs.append(user_attr_name)
                        user_module.attrs[f"{user_attr_name}_hash"] = Attr.STR()
                        session_auth_doc_args = cast(
                            MutableSequence[MutableMapping[str, "Attr"]],
                            Config.modules["session"].funcs["auth"].doc_attrs,
                        )
                        session_auth_doc_args.extend(
                            [
                                {
                                    "hash": Attr.STR(),
                                    user_attr_name: user_attr.type,
                                    "groups": Attr.LIST(list=[Attr.ID()]),
                                },
                                {"hash": Attr.STR(), user_attr_name: user_attr.type},
                            ]
                        )

                    if user_attr.default:
                        if not isinstance(user_attr.default, Default):
                            raise ConfigException(
                                f"Invalid type for attr default '{user_attr_name}' defined for "
                                f"package '{config.name}' for 'user_attrs' Config Attr"
                            )
                        user_module.defaults[user_attr_name] = user_attr.default

                    if user_attr.extn:
                        if not isinstance(user_attr.extn, Extn):
                            raise ConfigException(
                                f"Invalid type for attr extn '{user_attr_name}' defined for "
                                f"package '{config.name}' for 'user_attrs' Config Attr"
                            )
                        user_module.extns[user_attr_name] = user_attr.extn

                    if user_attr.func:
                        # [TODO] Expanded type validation for Func
                        if not isinstance(user_attr.type, Func):
                            raise ConfigException(
                                f"Invalid type for attr func '{user_attr_name}' defined for "
                                f"package '{config.name}' for 'user_attrs' Config Attr"
                            )
                        user_module.funcs[f"update_{user_attr_name}"] = user_attr.func
                    else:
                        user_module.funcs[f"update_{user_attr_name}"] = Func(
                            permissions=[Perm(privilege=user_attr_name)],
                            query_attrs={"_id": Attr.ID()},
                            doc_attrs={user_attr_name: user_attr.type},
                            strict_doc=True,
                        )

                    user_module.funcs[
                        f"update_{user_attr_name}"
                    ].name = f"update_{user_attr_name}"
                    user_module.funcs[f"update_{user_attr_name}"].module = user_module

            # For vars_types Config Attr, preserve config name for debugging purposes
            elif attr_name == "vars_types":
                for var in attr_val:
                    config.name = cast(str, config.name)
                    Config.vars_types[var] = {
                        "package": config.name,
                        "type": attr_val[var],
                    }

            # Otherwise, update Config accroding to attr_val type
            elif isinstance(attr_val, list):
                for j in attr_val:
                    getattr(Config, attr_name).append(j)
                if attr_name == "locales":
                    Config.locales = list(set(Config.locales))

            elif isinstance(attr_val, dict):
                if not getattr(Config, attr_name):
                    setattr(Config, attr_name, {})
                deep_update(target=getattr(Config, attr_name), new_values=attr_val)

            else:
                setattr(Config, attr_name, attr_val)

        if config.modules:
            config_modules = []
            for module in config.modules:
                if module.name in Config.modules:
                    raise ConfigException(
                        f"Module '{module.name}' exist in current runtime config"
                    )

                config_module(module_name=module.name, module=module)
                config_modules.append(module.name)
                Config.modules[module.name] = module

            Config.sys.packages[config.name or ""] = {
                "package": config,
                "modules": config_modules,
            }

        if getattr(config, "packages", None):
            pass

        if isinstance(config, App):
            Config.sys.name = config.name or ""
            Config.sys.version = config.version or ""

    # Update User, Session modules with populated attrs
    Config.modules["user"].defaults["locale"] = Default(value=Config.locale)

    # [DOC] Attempt to validate all packages required vars (via vars_types Config Attr) are met
    for var_name, var in Config.vars_types.items():
        if var_name not in Config.vars:
            raise ConfigException(
                f'Package \'{var["package"]}\' requires \'{var_name}\' Var, but not found in App '
                "Config"
            )
        try:
            validate_attr(
                mode="create",
                attr_name=var_name,
                attr_type=var["type"],
                attr_val=Config.vars[var_name],
            )
        except Exception as e:
            raise ConfigException(
                f'Package \'{var["package"]}\' requires \'{var_name}\' Var of type '
                f'\'{var["type"].type}\', but validation failed'
            ) from e

    # Validate Attr Type of type TYPE. Has to run before config_app so calls in config_app
    # depending on such attrs don't fail
    for attr_type in Config.sys.type_attrs:
        validate_type(attr_type=attr_type)

    # Attempt to call post_config callable of all modules
    for module_name, module in Config.modules.items():
        logger.debug(
            "Attempting to check for 'post_config' callable for module: %s",
            module_name,
        )
        if not module.post_config:
            logger.debug("- Module does not define 'post_config' callable")
            continue

        logger.debug("- Module does define 'post_config' callable. Calling it..")
        module.post_config()


def config_module(*, module_name: str, module: "Module"):
    """Validates args for 'module' and updates values for 'attrs' of 'module' per sets from
    'defaults', 'extns', 'counters'"""

    # pylint: disable=import-outside-toplevel

    import nawah.base as Base

    # Set name of module in Module object for reverse reference
    module.name = module_name

    logger.debug("Attempting to configure module '%s'", module_name)

    if not module.funcs:
        raise ConfigException(f"Module '{module_name}' doesn't define any function")

    for func_name, func in module.funcs.items():
        logger.debug(
            "Attempting to validate Function '%s'.'%s'", module_name, func_name
        )
        func.name = func_name
        func.module = module
        # Check value type
        if not isinstance(func, Func):
            raise ConfigException(
                f"Invalid Nawah Function for '{module_name}'.'{func_name}' of type "
                "'{type(func)}'"
            )

        # Validate query_attrs, doc_attrs, call_args
        for func_attr_name in ["query_attrs", "doc_attrs", "call_args"]:
            logger.debug(
                "Attempting to validate '%s' of '%s'.'%s'",
                func_attr_name,
                module_name,
                func_name,
            )
            if func_attr_val := getattr(func, func_attr_name):
                # query_attrs, doc_attrs are lists of dicts, call_args is just dict
                # Do this trick to get this loop to work nonetheless, by wrapping call_args with
                # a list, to let the rest of the loop deal with theree attrs in same manner
                if func_attr_name == "call_args":
                    func_attr_val = [func_attr_val]
                for attrs_set in func_attr_val:
                    for attr_name, attr_type in attrs_set.items():
                        try:
                            logger.debug(
                                "Attempting to validate '%s' Attr Type of '%s' for Function "
                                "'%s'.'%s' of set '%s'",
                                func_attr_name,
                                attr_name,
                                module_name,
                                func.name,
                                attrs_set,
                            )
                            validate_type(attr_type=attr_type, skip_type=True)
                        except InvalidAttrTypeException as e:
                            raise ConfigException(
                                f"Invalid Attr Type for '{module_name}'.'{attr_name}'. "
                                f"Original validation error: {str(e)}",
                            ) from e
                        except InvalidAttrTypeArgException as e:
                            raise ConfigException(
                                f"Invalid Attr Type Arg for '{module_name}'.'{attr_name}'. "
                                f"Original validation error: {str(e)}"
                            ) from e

        # [TODO] Add steps to update Func object attributes per func callable
        # If Func has callable defined skip
        if func.callable:
            if not callable(func.callable):
                raise ConfigException(
                    f"Value 'callable' for Nawah Function '{func_name}' of '{module_name}' "
                    "is not callable"
                )
            continue

        logger.debug(
            "Function '%s'.'%s' doesn't define callable. Attempting to set per name",
            module_name,
            func_name,
        )
        # Otherwise, set callable based on fucntion name, update query_attrs, doc_attrs, call_args
        if func_name.startswith("create_file"):
            func.doc_attrs = [
                {
                    "attr": Attr.STR(),
                    "lastModified": Attr.INT(),
                    "type": Attr.STR(),
                    "name": Attr.STR(),
                    "content": Attr(
                        desc="",
                        type=AttrType.BYTES,
                        args={},
                    ),
                },
                {"_id": Attr.ID(), "doc": Attr.ID()},
            ]
            func.post_func = True
            func.callable = Base.create_file

        elif func_name.startswith("update_file"):
            func.post_func = True
            func.callable = Base.update_file

        elif func_name.startswith("delete_file"):
            func.post_func = True
            func.callable = Base.delete_file

        elif func_name.startswith("delete_lock"):
            func.callable = Base.delete_lock

        elif func_name.startswith("obtain_lock"):
            func.callable = Base.obtain_lock

        elif func_name.startswith("read_diff"):
            func.callable = Base.read_diff

        elif func_name.startswith("retrieve_file"):
            func.query_attrs = [
                {
                    "_id": Attr.ID(),
                    "attr": Attr.STR(),
                    "filename": Attr.STR(),
                },
                {
                    "_id": Attr.ID(),
                    "attr": Attr.STR(),
                    "thumb": Attr.STR(pattern=r"[0-9]+x[0-9]+"),
                    "filename": Attr.STR(),
                },
            ]
            func.get_func = True
            func.callable = Base.retrieve_file

        elif func_name.startswith("read"):
            func.callable = Base.read

        elif func_name.startswith("create"):
            func.callable = Base.create

        elif func_name.startswith("update"):
            func.callable = Base.update

        elif func_name.startswith("delete"):
            func.callable = Base.delete

        else:
            raise ConfigException(
                f"Can't set callable for Nawah Function '{func_name}' of"
                f" '{module_name}'"
            )

        if not func.call_args:
            func.call_args = {}

        func.call_args.update({"raise_no_success": Attr.BOOL()})

    # Check attrs for any invalid type
    for attr in module.attrs:
        try:
            logger.debug(
                "Attempting to validate Attr Type for '%s'.'%s'",
                module_name,
                attr,
            )
            validate_type(attr_type=module.attrs[attr], skip_type=True)
        except InvalidAttrTypeException as e:
            raise ConfigException(
                f"Invalid Attr Type for '{module_name}'.'{attr}'. Original "
                f"validation error: {str(e)}",
            ) from e
        except InvalidAttrTypeArgException as e:
            raise ConfigException(
                f"Invalid Attr Type Arg for '{module_name}'.'{attr}'. Original "
                f"validation error: {str(e)}"
            ) from e

    # Check defaults for invalid types, update default value
    for attr_name, default in module.defaults.items():
        if not isinstance(default, Default):
            raise ConfigException(
                f"Invalid Default Instruction for '{module_name}'.'{attr_name}', of type "
                f"'{type(default)}'"
            )

        logger.debug(
            "Updating default value for attr '%s' to: '%s'",
            attr_name,
            module.defaults[attr_name],
        )
        extract_attr(attrs=module.attrs, path=attr_name).default = module.defaults[
            attr_name
        ]

    # Check counters for invalid attrs, types
    for attr_name, counter in module.counters.items():
        if not isinstance(counter, Counter):
            raise ConfigException(
                f"Invalid Counter Instruction for '{module_name}'.'{attr_name}', of type "
                f"'{type(counter)}'"
            )

        if attr_name not in module.attrs:
            raise ConfigException(
                f"Invalid attr for Counter Instruction. Module '{module_name}' "
                f"doesn't have top-level attr '{attr_name}'"
            )

        if module.attrs[attr_name].type != AttrType.ANY:
            raise ConfigException(
                f"Attr '{module_name}'.'{attr_name}' for Counter Instruction can only be of "
                "Attr Type ANY"
            )

        # Set default value to Default(value=None) to skip InvalidAttrException when checking doc
        module.attrs[attr_name].default = Default(value=None)

    # Update extn value
    for attr_name, extn in module.extns.items():
        if not isinstance(extn, Extn):
            raise ConfigException(
                f"Invalid Extension Instruction for '{module_name}'.'{attr_name}', of type "
                f"'{type(extn)}'"
            )

        logger.debug(
            "Updating 'extn' for attr '%s' to: '%s'",
            attr_name,
            module.extns[attr_name],
        )
        extract_attr(attrs=module.attrs, path=attr_name).extn = module.extns[attr_name]

    # Check extns for invalid extended attrs
    for attr_name, extn in module.extns.items():
        if not isinstance(extn, Extn):
            raise ConfigException(
                f"Invalid extns attr '{attr_name}' of module '{module_name}'"
            )

    # Check valid type, value for diff
    if module.diff is not None and not isinstance(module.diff, Diff):
        raise ConfigException(f"Invalid Diff Instruction for module '{module_name}'")

    if isinstance(module.diff, Attr):
        module.diff = cast(Attr, module.diff)
        if module.diff.type != AttrType.TYPE:
            raise ConfigException(
                f"Invalid Attr Type for diff of module '{module_name}'. Only Attr Type TYPE "
                "is allowed"
            )
        logger.debug(
            "Attempting to validate Attr Type diff of module '%s'", module_name
        )
        try:
            validate_type(attr_type=module.diff)
        except InvalidAttrTypeException as e:
            raise ConfigException(
                f"Invalid Attr Type for diff of module '{module_name}'. Original validation error: {str(e)}"
            ) from e
        except InvalidAttrTypeArgException as e:
            raise ConfigException(
                f"Invalid Attr Type Arg for diff of module '{module_name}'. Original validation error: {str(e)}"
            ) from e

    # [TODO] Re-implement
    # Check valid types, values for create_draft, update_draft
    # for attr in ['create_draft', 'update_draft']:
    #     if type(getattr(self, attr)) not in [bool, Attr]:
    #         raise ConfigException(f'Invalid {attr} for module \'{module_name}\'') from e
    #     if type(getattr(self, attr)) == Attr:
    #         if getattr(self, attr).type != AttrType.TYPE:
    #             raise ConfigException(
    #                 f'Invalid Attr Type for {attr} of module \'{module_name}\'. Only Attr Type TYPE is allowed'
    #             ) from e
    #         logger.debug(
    #             f'Attempting to validate Attr Type {attr} of module \'{module_name}\'.'
    #         )
    #         try:
    #             validate_type(attr_type=getattr(self, attr))
    #         except InvalidAttrTypeException as e:
    #             raise ConfigException(
    #                 f'Invalid Attr Type for {attr} of module \'{module_name}\'. Original validation error: {str(e)}'
    #             ) from e
    #         except InvalidAttrTypeArgException as e:
    #             raise ConfigException(
    #                 f'Invalid Attr Type Arg for {attr} of module \'{module_name}\'. Original validation error: {str(e)}'
    #             ) from e

    logger.debug("Configured module %s", module_name)


async def config_app():

    # [TODO] Re-implement
    # Check API version
    # nawah_level = '.'.join(Config.sys.version.split('.')[0:2])
    # for package, api_level in {
    #     name: config['package'].api_level
    #     for name, config in Config.sys.packages.items()
    # }.items():
    #     if api_level != nawah_level:
    #         raise ConfigException(
    #             f'Nawah framework is on API-level \'{nawah_level}\', but the app package'
    #             f'\'{package}\' requires API-level \'{api_level}\''
    #         )
    # try:
    #     versions = (
    #         (
    #             requests.get(
    #                 'https://raw.githubusercontent.com/masaar/nawah_versions/master/versions.txt'
    #             ).content
    #         )
    #         .decode('utf-8')
    #         .split('\n')
    #     )
    #     version_detected = ''
    #     for version in versions:
    #         if version.startswith(f'{nawah_level}.'):
    #             if version_detected and int(version.split('.')[-1]) < int(
    #                 version_detected.split('.')[-1]
    #             ):
    #                 continue
    #             version_detected = version
    #     if version_detected and version_detected != Config.sys.version:
    #         logger.warning(
    #             f'Your app is using Nawah version \'{Config.sys.version}\' while newer version \'{version_detected}\' of the API-level is available. Please, update'
    #         )
    # except:
    #     logger.warning(
    #         'An error occurred while attempting to check for latest update to Nawah. Please, check for updates on your own'
    #     )

    # Check for jobs
    if Config.jobs:
        # Check jobs schedule validity
        Config.sys.jobs_base = datetime.datetime.utcnow()
        for job_name, job in Config.jobs.items():
            if not croniter.is_valid(job.schedule):
                raise ConfigException(
                    f"Job with schedule '{job_name}' schedule is invalid"
                )

            job.cron_schedule = croniter(job.schedule, Config.sys.jobs_base)
            job.next_time = datetime.datetime.fromtimestamp(
                job.cron_schedule.get_next(), datetime.timezone.utc
            ).isoformat()[:16]

    # Check for presence of user_attrs
    if not Config.user_attrs:
        raise ConfigException("No 'user_attrs' are provided")

    # Check for Env Vars
    attrs_defaults = {
        "data_server": "mongodb://localhost",
        "data_name": "nawah_data",
        "data_ssl": False,
        "emulate_test": False,
    }
    for attr_name in attrs_defaults:
        attr_val = getattr(Config, attr_name)
        if type(attr_val) == str and attr_val.startswith("$__env."):
            logger.debug(f"Detected Env Variable for config attr '{attr_name}'")
            if not os.getenv(attr_val[7:]):
                logger.warning(
                    f"Couldn't read Env Variable for config attr '{attr_name}'. Defaulting to '{attrs_defaults[attr_name]}'"
                )
                setattr(Config, attr_name, attrs_defaults[attr_name])
            else:
                # Set data_ssl to True rather than string Env Variable value
                if attr_name == "ssl":
                    attr_val = True
                else:
                    attr_val = os.getenv(attr_val[7:])
                logger.warning(
                    f"Setting Env Variable for config attr '{attr_name}' to '{attr_val}'"
                )
                setattr(Config, attr_name, attr_val)

    from nawah import data as Data

    # Create default env dict
    anon_user = _compile_anon_user()
    anon_session = _compile_anon_session()
    anon_session["user"] = anon_user
    Config.sys.conn = Data.create_conn()
    Config.sys.env = {
        "conn": Config.sys.conn,
        "REMOTE_ADDR": "127.0.0.1",
        "HTTP_USER_AGENT": "Nawah",
        "client_app": "__sys",
        "session": anon_session,
    }

    # Check cache_server Config Attr
    if Config.cache_server:
        logger.debug("Attempting to create connection with 'cache_server'")
        Config.sys.cache = aioredis.Redis(
            host=Config.cache_server,
            db=Config.cache_db,
            username=Config.cache_username,
            password=Config.cache_password,
        ).json()

    # Check error_reporting_server Config Attr
    if Config.error_reporting_server:
        logger.debug("Attempting to create connection with 'error_reporting_server'")
        sentry_sdk.init(  # pylint: disable=abstract-class-instantiated
            Config.error_reporting_server,
            traces_sample_rate=1.0,
            with_locals=True,
            release=f"{Config.sys.name}:{Config.sys.version}",
            environment=Config.env,
        )

    # Check test mode
    if Config.test:
        logger.debug("Test mode detected")
        logger.setLevel(logging.DEBUG)
        __location__ = os.path.realpath(os.path.join("."))
        if not os.path.exists(os.path.join(__location__, "tests")):
            os.makedirs(os.path.join(__location__, "tests"))
        for module_name, module in Config.modules.items():
            module_collection = module.collection
            logger.debug(
                f"Updating collection name '{module_collection}' of module {module_name}"
            )
            module_collection = module.collection = f"test_{module_collection}"

            logger.debug(f"Flushing test collection '{module_collection}'")
            await Data.drop(
                env=Config.sys.env,
                collection_name=module_collection,
            )

    # Checking users collection
    # [TODO] Updated sequence to handle users
    logger.debug("Testing users collection")
    if Config.sys.admin_check:
        user_results = await call(
            "user/read",
            # [TODO] Check behaviour with removed Event.ON
            skip_events=[Event.PERM],
            env=Config.sys.env,
            query=[{"_id": "f00000000000000000000010"}],
        )
        if not user_results["args"]["count"]:
            logger.debug("ADMIN user not found, creating it")
            # Prepare base ADMIN user doc
            admin_create_doc = {
                "_id": ObjectId("f00000000000000000000010"),
                "name": {Config.locale: "__ADMIN"},
                "groups": [],
                "privileges": {"*": ["*"]},
                "locale": Config.locale,
            }
            # Update ADMIN user doc with admin_doc Config Attr
            admin_create_doc.update(Config.admin_doc)

            for auth_attr in Config.user_attrs:
                admin_create_doc[f"{auth_attr}_hash"] = pbkdf2_sha512.using(
                    rounds=100000
                ).hash(
                    f"{auth_attr}{admin_create_doc[auth_attr]}{Config.admin_password}"
                    f"{Config.anon_token}".encode("utf-8")
                )
            admin_results = await call(
                "user/create",
                skip_events=[Event.PERM],
                env=Config.sys.env,
                doc=admin_create_doc,
            )
            logger.debug("ADMIN user creation results: %s", admin_results)
            if admin_results["status"] != 200:
                raise ConfigException("Config step failed")
        else:
            logger.debug(
                "'ADMIN' doc found. Attempting to check if update is required.."
            )
            admin_doc = user_results["args"]["docs"][0]
            admin_doc_update = {}
            for attr_name, attr_val in Config.admin_doc.items():
                if (
                    attr_name not in admin_doc
                    or not admin_doc[attr_name]
                    or attr_val != admin_doc[attr_name]
                ):
                    if (
                        isinstance(attr_val, dict)
                        and Config.locale in attr_val
                        and isinstance(admin_doc[attr_name], dict)
                        and (
                            (
                                Config.locale in admin_doc[attr_name]
                                and attr_val[Config.locale]
                                == admin_doc[attr_name][Config.locale]
                            )
                            or (Config.locale not in admin_doc[attr_name])
                        )
                    ):
                        continue
                    logger.debug(
                        "Detected change in 'admin_doc.%s' Config Attr", attr_name
                    )
                    admin_doc_update[attr_name] = attr_val
            for auth_attr in Config.user_attrs:
                auth_attr_hash = pbkdf2_sha512.using(rounds=100000).hash(
                    f"{auth_attr}{admin_doc[auth_attr]}{Config.admin_password}"
                    f"{Config.anon_token}".encode("utf-8")
                )
                if (
                    f"{auth_attr}_hash" not in admin_doc
                    or auth_attr_hash != admin_doc[f"{auth_attr}_hash"]
                ):
                    logger.debug(f"Detected change in 'admin_password' Config Attr")
                    admin_doc_update[f"{auth_attr}_hash"] = auth_attr_hash
            if len(admin_doc_update.keys()):
                logger.debug(
                    f"Attempting to update ADMIN user with doc: '{admin_doc_update}'"
                )
                admin_results = await call(
                    "user/update",
                    # [TODO] Check behaviour with removed Event.ON
                    # [TODO] Check behaviour with removed Event.PRE
                    skip_events=[Event.PERM],
                    env=Config.sys.env,
                    query=[{"_id": ObjectId("f00000000000000000000010")}],
                    doc=admin_doc_update,
                )
                logger.debug(f"ADMIN user update results: {admin_results}")
                if admin_results["status"] != 200:
                    raise ConfigException("Config step failed")
            else:
                logger.debug("ADMIN user is up-to-date")
    else:
        logger.warning(
            "Skipping checking 'ADMIN' doc due to 'sys.admin_check' Config Attr"
        )

    Config.sys.docs[ObjectId("f00000000000000000000010")] = SysDoc(module="user")

    # Test if ANON user exists
    user_results = await call(
        "user/read",
        # [TODO] Check behaviour with removed Event.ON
        skip_events=[Event.PERM],
        env=Config.sys.env,
        query=[{"_id": "f00000000000000000000011"}],
    )
    if not user_results["args"]["count"]:
        logger.debug("ANON user not found, creating it")
        anon_results = await call(
            "user/create",
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM],
            env=Config.sys.env,
            doc=_compile_anon_user(),
        )
        logger.debug(f"ANON user creation results: {anon_results}")
        if anon_results["status"] != 200:
            raise ConfigException("Config step failed")
    else:
        logger.debug("ANON user found, checking it")
        anon_doc = _compile_anon_user()
        anon_doc_update = {}
        for attr in Config.user_attrs:
            if attr not in anon_doc or not anon_doc[attr]:
                logger.debug(f"Detected change in 'anon_doc.{attr}' Config Attr")
                anon_doc_update[attr] = generate_attr_val(
                    attr_type=Config.user_attrs[attr].type
                )
        for module in Config.anon_privileges:
            if module not in anon_doc or set(anon_doc[module]) != set(
                Config.anon_privileges[module]
            ):
                logger.debug(f"Detected change in 'anon_privileges' Config Attr")
                anon_doc_update[f"privileges.{module}"] = Config.anon_privileges[module]
        for auth_attr in Config.user_attrs:
            if (
                f"{auth_attr}_hash" not in anon_doc
                or anon_doc[f"{auth_attr}_hash"] != Config.anon_token
            ):
                logger.debug(f"Detected change in 'anon_token' Config Attr")
                anon_doc_update[attr] = Config.anon_token
            anon_doc_update[f"{auth_attr}_hash"] = Config.anon_token
        if len(anon_doc_update.keys()):
            logger.debug(
                f"Attempting to update ANON user with doc: '{anon_doc_update}'"
            )
            anon_results = await call(
                "user/update",
                # [TODO] Check behaviour with removed Event.ON
                # [TODO] Check behaviour with removed Event.PRE
                skip_events=[Event.PERM],
                env=Config.sys.env,
                query=[{"_id": ObjectId("f00000000000000000000011")}],
                doc=anon_doc_update,
            )
            logger.debug(f"ANON user update results: {anon_results}")
            if anon_results["status"] != 200:
                raise ConfigException("Config step failed")
        else:
            logger.debug("ANON user is up-to-date")

    Config.sys.docs[ObjectId("f00000000000000000000011")] = SysDoc(module="user")

    logger.debug("Testing sessions collection")
    # Test if ANON session exists
    session_results = await call(
        "session/read",
        # [TODO] Check behaviour with removed Event.ON
        skip_events=[Event.PERM],
        env=Config.sys.env,
        query=[{"_id": "f00000000000000000000012"}],
    )
    if not session_results["args"]["count"]:
        logger.debug("ANON session not found, creating it")
        anon_results = await call(
            "session/create",
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM],
            env=Config.sys.env,
            doc=_compile_anon_session(),
        )
        logger.debug(f"ANON session creation results: {anon_results}")
        if anon_results["status"] != 200:
            raise ConfigException("Config step failed")
    Config.sys.docs[ObjectId("f00000000000000000000012")] = SysDoc(module="session")

    logger.debug("Testing groups collection")
    # Test if DEFAULT group exists
    group_results = await call(
        "group/read",
        # [TODO] Check behaviour with removed Event.ON
        skip_events=[Event.PERM],
        env=Config.sys.env,
        query=[{"_id": "f00000000000000000000013"}],
    )
    if not group_results["args"]["count"]:
        logger.debug("DEFAULT group not found, creating it")
        group_create_doc = {
            "_id": ObjectId("f00000000000000000000013"),
            "user": ObjectId("f00000000000000000000010"),
            "name": {locale: "__DEFAULT" for locale in Config.locales},
            "bio": {locale: "__DEFAULT" for locale in Config.locales},
            "privileges": Config.default_privileges,
        }
        group_results = await call(
            "group/create",
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM],
            env=Config.sys.env,
            doc=group_create_doc,
        )
        logger.debug("DEFAULT group creation results: %s", group_results)
        if group_results["status"] != 200:
            raise ConfigException("Config step failed")
    else:
        logger.debug("DEFAULT group found, checking it")
        default_doc = group_results["args"]["docs"][0]
        default_doc_update: MutableMapping[str, Any] = {}
        for module in Config.default_privileges:
            if module not in default_doc["privileges"].keys() or set(
                default_doc["privileges"][module]
            ) != set(Config.default_privileges[module]):
                logger.debug("Detected change in 'default_privileges' Config Attr")
                default_doc_update[f"privileges.{module}"] = Config.default_privileges[
                    module
                ]
        if len(default_doc_update.keys()):
            logger.debug(
                "Attempting to update DEFAULT group with doc: '%s'",
                default_doc_update,
            )
            default_results = await call(
                "group/update",
                # [TODO] Check behaviour with removed Event.ON
                # [TODO] Check behaviour with removed Event.PRE
                skip_events=[Event.PERM],
                env=Config.sys.env,
                query=[{"_id": ObjectId("f00000000000000000000013")}],
                doc=default_doc_update,
            )
            logger.debug(f"DEFAULT group update results: {default_results}")
            if anon_results["status"] != 200:
                raise ConfigException("Config step failed")
        else:
            logger.debug("DEFAULT group is up-to-date")

    Config.sys.docs[ObjectId("f00000000000000000000013")] = SysDoc(module="group")

    # Test app-specific groups
    logger.debug("Testing app-specific groups collection")
    for group in Config.groups:
        group_results = await call(
            "group/read",
            # [TODO] Check behaviour with removed Event.ON
            skip_events=[Event.PERM],
            env=Config.sys.env,
            query=[{"_id": group["_id"]}],
        )
        if not group_results["args"]["count"]:
            logger.debug(
                f'App-specific group with name \'{group["name"]}\' not found, creating it'
            )
            group_results = await call(
                "group/create",
                # [TODO] Check behaviour with removed Event.ON
                # [TODO] Check behaviour with removed Event.PRE
                skip_events=[Event.PERM],
                env=Config.sys.env,
                doc=group,
            )
            logger.debug(
                f'App-specific group with name {group["name"]} creation results: {group_results}'
            )
            if group_results["status"] != 200:
                raise ConfigException("Config step failed")
        else:
            logger.debug(
                f'App-specific group with name \'{group["name"]}\' found, checking it'
            )
            group_doc = group_results["args"]["docs"][0]
            group_doc_update = {}
            if "privileges" in group:
                for module in group["privileges"]:
                    if module not in group_doc["privileges"].keys() or set(
                        group_doc["privileges"][module]
                    ) != set(group["privileges"][module]):
                        logger.debug(
                            f'Detected change in \'privileges\' Doc Arg for group with name \'{group["name"]}\''
                        )
                        group_doc_update[f"privileges.{module}"] = group["privileges"][
                            module
                        ]
            if len(group_doc_update.keys()):
                logger.debug(
                    "Attempting to update group with name '%s' with doc: '%s'",
                    group["name"],
                    group_doc_update,
                )
                group_results = await call(
                    "group/update",
                    # [TODO] Check behaviour with removed Event.ON
                    # [TODO] Check behaviour with removed Event.PRE
                    skip_events=[Event.PERM],
                    env=Config.sys.env,
                    query=[{"_id": group["_id"]}],
                    doc=group_doc_update,
                )
                logger.debug(
                    "Group with name '%s' update results: %s",
                    group["name"],
                    group_results,
                )
                if group_results["status"] != 200:
                    raise ConfigException("Config step failed")
            else:
                logger.debug("Group with name '%s' is up-to-date", group["name"])

        Config.sys.docs[ObjectId(group["_id"])] = SysDoc(module="group")

    # Test app-specific data indexes
    logger.debug("Testing data indexes")
    for index in Config.data_indexes:
        logger.debug("Attempting to create data index: %s", index)
        try:
            Config.sys.conn[Config.data_name][index["collection"]].create_index(
                index["index"]
            )
        except Exception as e:
            logger.error("Failed to create data index: %s, with error: %s", index, e)
            logger.error("Evaluate error and take action manually")

    logger.debug("Creating 'var', 'type', 'user' data indexes for settings collections")
    Config.sys.conn[Config.data_name]["settings"].create_index([("var", 1)])
    Config.sys.conn[Config.data_name]["settings"].create_index([("type", 1)])
    Config.sys.conn[Config.data_name]["settings"].create_index([("user", 1)])
    logger.debug(
        "Creating 'user', 'event', 'subevent' data indexes for analytics collections"
    )
    Config.sys.conn[Config.data_name]["analytics"].create_index([("user", 1)])
    Config.sys.conn[Config.data_name]["analytics"].create_index([("event", 1)])
    Config.sys.conn[Config.data_name]["analytics"].create_index([("subevent", 1)])
    logger.debug("Creating '__deleted' data indexes for all collections")
    for module in Config.modules:
        if Config.modules[module].collection:
            logger.debug(
                "Attempting to create '__deleted' data index for collection: %s",
                Config.modules[module].collection,
            )
            Config.sys.conn[Config.data_name][
                Config.modules[module].collection
            ].create_index([("__deleted", 1)])

    # Test app-specific docs
    logger.debug("Testing docs")
    for doc in Config.docs:
        if not isinstance(doc, SysDoc):
            raise ConfigException("Invalid Config Attr 'docs'")

        doc_results = await call(
            f"{doc.module}/read",
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM, Event.ATTRS_QUERY],
            env=Config.sys.env,
            query=[{doc.key: doc.key_value}],  # type: ignore
        )
        if not doc_results["args"]["count"]:
            skip_events = [Event.PERM]
            if doc.skip_args is True:
                skip_events.append(Event.ATTRS_DOC)
            doc.doc = cast("NawahDoc", doc.doc)
            doc_results = await call(
                f"{doc.module}/create",
                skip_events=skip_events,
                env=Config.sys.env,
                doc=doc.doc,
            )
            logger.debug(
                "App-specific doc with %s '%s' of module '%s' creation results: %s",
                doc.key,
                doc.key_value,
                doc.module,
                doc_results,
            )
            if doc_results["status"] != 200:
                raise ConfigException("Config step failed")
        Config.sys.docs[ObjectId(doc_results["args"]["docs"][0]["_id"])] = SysDoc(
            module=doc.module
        )

    # Check for emulate_test mode
    if Config.emulate_test:
        Config.test = True


def _compile_anon_user():
    anon_doc = {
        "_id": ObjectId("f00000000000000000000011"),
        "name": {Config.locale: "__ANON"},
        "groups": [],
        "privileges": Config.anon_privileges,
        "locale": Config.locale,
    }
    for user_attr_name, user_attr in Config.user_attrs.items():
        if user_attr.default:
            anon_doc[user_attr_name] = copy.deepcopy(user_attr.default.value)
        else:
            anon_doc[user_attr_name] = generate_attr_val(attr_type=user_attr.type)
    for auth_attr in Config.auth_attrs:
        anon_doc[f"{auth_attr}_hash"] = Config.anon_token
    if Config.anon_doc:
        anon_doc.update(Config.anon_doc)
    return anon_doc


def _compile_anon_session():
    session_doc = {
        "_id": ObjectId("f00000000000000000000012"),
        "user": ObjectId("f00000000000000000000011"),
        "host_add": "127.0.0.1",
        "user_agent": Config.anon_token,
        "timestamp": "1970-01-01T00:00:00",
        "expiry": "1970-01-01T00:00:00",
        "token": Config.anon_token,
        "token_hash": Config.anon_token,
    }
    return session_doc



"""Provides exceptions related to Query"""

from typing import Any, Literal


class InvalidQueryStepTypeException(Exception):
    """Raised by \'validate_query_step\' if type of Query step is not 'dict'"""

    status = 400

    def __init__(self, *, step_type: Any):
        super().__init__(
            InvalidQueryStepTypeException.format_msg(step_type=step_type),
            {"step_type": step_type},
        )

    @staticmethod
    def format_msg(*, step_type):
        """Formats exception message"""

        return f"Query step type should be of type 'dict'. Got '{step_type}' instead"


class InvalidQueryStepLenException(Exception):
    """Raised by \'validate_query_step\' if Query step has more than one item"""

    status = 400

    def __init__(self, *, step_items: tuple[str]):
        super().__init__(
            InvalidQueryStepLenException.format_msg(step_items=step_items),
            {"step_items": step_items},
        )

    @staticmethod
    def format_msg(*, step_items):
        """Formats exception message"""

        return f"Query step can only have one item. Got '{step_items}' instead"


class InvalidQueryStepAttrTypeException(Exception):
    """Raised by \'validate_query_step\' if type of Query step attr is not 'dict'"""

    status = 400

    def __init__(self, *, step_attr_type: Any):
        super().__init__(
            InvalidQueryStepAttrTypeException.format_msg(step_attr_type=step_attr_type),
            {"step_attr_type": step_attr_type},
        )

    @staticmethod
    def format_msg(*, step_attr_type):
        """Formats exception message"""

        return f"Query step attr value should be of type 'dict'. Got '{step_attr_type}' instead"


class InvalidQueryStepAttrLenException(Exception):
    """Raised by \'validate_query_step\' if Query step attr has more than one item"""

    status = 400

    def __init__(self, *, step_attr_items: tuple[str]):
        super().__init__(
            InvalidQueryStepAttrLenException.format_msg(
                step_attr_items=step_attr_items
            ),
            {"step_attr_items": step_attr_items},
        )

    @staticmethod
    def format_msg(*, step_attr_items):
        """Formats exception message"""

        return (
            f"Query step attr can only have one item. Got '{step_attr_items}' instead"
        )


class InvalidQueryOperTypeException(Exception):
    """Raised by \'validate_query_step\' if value of Query step attr is invalid"""

    status = 400

    def __init__(
        self,
        *,
        attr_name: str,
        attr_oper: Literal[
            "$ne",
            "$eq",
            "$gt",
            "$gte",
            "$lt",
            "$lte",
            "$all",
            "$in",
            "$nin",
            "$regex",
        ],
        attr_type: Any,
        attr_val: Any,
    ):
        super().__init__(
            InvalidQueryOperTypeException.format_msg(
                attr_name=attr_name,
                attr_oper=attr_oper,
                attr_type=attr_type,
                attr_val=attr_val,
            ),
            {
                "attr_name": attr_name,
                "attr_oper": attr_oper,
                "attr_type": attr_type,
                "attr_val": attr_val,
            },
        )

    @staticmethod
    def format_msg(*, attr_name, attr_oper, attr_type, attr_val):
        """Formats exception message"""

        return f"Invalid value for Query Arg '{attr_name}' with Query Arg Oper '{attr_oper}' expecting type '{attr_type}' but got '{attr_val}'"


class UnknownQueryOperException(Exception):
    """Raised by \'validate_query_step\' if an unknown \'Query Arg\' is detected"""

    status = 400

    def __init__(
        self,
        *,
        attr_name: str,
        attr_oper: Literal[
            "$ne",
            "$eq",
            "$gt",
            "$gte",
            "$lt",
            "$lte",
            "$all",
            "$in",
            "$nin",
            "$regex",
        ],
    ):
        super().__init__(
            UnknownQueryOperException.format_msg(
                attr_name=attr_name, attr_oper=attr_oper
            ),
            {"attr_name": attr_name, "attr_oper": attr_oper},
        )

    @staticmethod
    def format_msg(*, attr_name, attr_oper):
        """Formats exception message"""

        return f"Unknown Query Arg Oper '{attr_oper}' for Query Arg '{attr_name}'"
